"""Classes for QML and JavaScript languages."""

from __future__ import annotations

from prettyqt.qt.QtQml import *  # noqa: F403

from prettyqt.qt.QtQml import (
    qmlRegisterType as register_qml_type,  # noqa: N813
    qmlClearTypeRegistrations as clear_type_registrations,  # noqa: N813
)
from .qmlparserstatus import QmlParserStatus, QmlParserStatusMixin
from .qmlerror import QmlError
from .qmlexpression import QmlExpression
from .qmlproperty import QmlProperty
from .qmlpropertymap import QmlPropertyMap
from .jsvalue import JSValue
from .jsvalueiterator import JSValueIterator
from .jsengine import JSEngine, JSEngineMixin
from .qmlengine import QmlEngine, QmlEngineMixin
from .qmlapplicationengine import QmlApplicationEngine
from .qmlcomponent import QmlComponent
from .qmlimageproviderbase import QmlImageProviderBase, QmlImageProviderBaseMixin
from prettyqt.qt import QtQml

QT_MODULE = QtQml


def register_objects_from_module(module):
    from prettyqt import core
    from prettyqt.utils import classhelpers

    for kls in classhelpers.iter_classes_for_module(module):
        if core.ObjectMixin in kls.mro():
            register_class(kls)


def register_class(kls):
    register_qml_type(kls, kls.__module__, 1, 0, kls.__name__)


__all__ = [
    "JSEngine",
    "JSEngineMixin",
    "JSValue",
    "JSValueIterator",
    "QmlApplicationEngine",
    "QmlComponent",
    "QmlEngine",
    "QmlEngineMixin",
    "QmlError",
    "QmlExpression",
    "QmlImageProviderBase",
    "QmlImageProviderBaseMixin",
    "QmlParserStatus",
    "QmlParserStatusMixin",
    "QmlProperty",
    "QmlPropertyMap",
    "clear_type_registrations",
    "register_qml_type",
]



from __future__ import annotations

from prettyqt import charts


class BarCategoryAxis(charts.AbstractAxisMixin, charts.QBarCategoryAxis):
    def __delitem__(self, index: str):
        """Remove category."""
        self.remove(index)

    def __getitem__(self, index: int) -> str:
        """Return category for given index."""
        return self.categories()[index]

    def __setitem__(self, index: str, value: str):
        """Set category at given index to value."""
        self.replace(index, value)

    def __add__(self, other: str) -> BarCategoryAxis:
        """Append another category to axis."""
        self.append(other)
        return self



from __future__ import annotations

import contextlib

from prettyqt import core


class BaseEventFilter(core.Object):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.objectName():
            self.setObjectName(type(self).__name__)

    @contextlib.contextmanager
    def applied_to(self, obj):
        obj.installEventFilter(self)
        yield obj
        obj.removeEventFilter(self)


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    widget = widgets.Widget()
    eventfilter = BaseEventFilter()
    with eventfilter.applied_to(widget):
        pass



from __future__ import annotations

from typing import Literal, Self

from prettyqt import constants
from prettyqt.qt import QtCore
from prettyqt.utils import bidict, get_repr


Loc = QtCore.QLocale
CN = Loc.Country

COUNTRY = bidict(
    any=CN.AnyCountry,
    afghanistan=CN.Afghanistan,
    aland_islands=CN.AlandIslands,
    albania=CN.Albania,
    algeria=CN.Algeria,
    american_samoa=CN.AmericanSamoa,
    andorra=CN.Andorra,
    angola=CN.Angola,
    anguilla=CN.Anguilla,
    antarctica=CN.Antarctica,
    antigua_and_barbuda=CN.AntiguaAndBarbuda,
    argentina=CN.Argentina,
    armenia=CN.Armenia,
    aruba=CN.Aruba,
    ascension_island=CN.AscensionIsland,
    australia=CN.Australia,
    austria=CN.Austria,
    azerbaijan=CN.Azerbaijan,
    bahamas=CN.Bahamas,
    bahrain=CN.Bahrain,
    bangladesh=CN.Bangladesh,
    barbados=CN.Barbados,
    belarus=CN.Belarus,
    belgium=CN.Belgium,
    belize=CN.Belize,
    benin=CN.Benin,
    bermuda=CN.Bermuda,
    bhutan=CN.Bhutan,
    bolivia=CN.Bolivia,
    bonaire=CN.Bonaire,
    bosnia_and_herzegowina=CN.BosniaAndHerzegowina,
    botswana=CN.Botswana,
    bouvet_island=CN.BouvetIsland,
    brazil=CN.Brazil,
    british_indian_ocean_territory=CN.BritishIndianOceanTerritory,
    british_virgin_islands=CN.BritishVirginIslands,
    brunei=CN.Brunei,
    bulgaria=CN.Bulgaria,
    burkina_faso=CN.BurkinaFaso,
    burundi=CN.Burundi,
    cambodia=CN.Cambodia,
    cameroon=CN.Cameroon,
    canada=CN.Canada,
    canary_islands=CN.CanaryIslands,
    cape_verde=CN.CapeVerde,
    cayman_islands=CN.CaymanIslands,
    central_african_republic=CN.CentralAfricanRepublic,
    ceuta_and_melilla=CN.CeutaAndMelilla,
    chad=CN.Chad,
    chile=CN.Chile,
    china=CN.China,
    christmas_island=CN.ChristmasIsland,
    clipperton_island=CN.ClippertonIsland,
    cocos_islands=CN.CocosIslands,
    colombia=CN.Colombia,
    comoros=CN.Comoros,
    congo_brazzaville=CN.CongoBrazzaville,
    congo_kinshasa=CN.CongoKinshasa,
    cook_islands=CN.CookIslands,
    costa_rica=CN.CostaRica,
    croatia=CN.Croatia,
    cuba=CN.Cuba,
    curasao=CN.CuraSao,
    cyprus=CN.Cyprus,
    czech_republic=CN.CzechRepublic,
    denmark=CN.Denmark,
    diego_garcia=CN.DiegoGarcia,
    djibouti=CN.Djibouti,
    dominica=CN.Dominica,
    dominican_republic=CN.DominicanRepublic,
    east_timor=CN.EastTimor,
    ecuador=CN.Ecuador,
    egypt=CN.Egypt,
    el_salvador=CN.ElSalvador,
    equatorial_guinea=CN.EquatorialGuinea,
    eritrea=CN.Eritrea,
    estonia=CN.Estonia,
    ethiopia=CN.Ethiopia,
    european_union=CN.EuropeanUnion,
    europe=CN.Europe,
    falkland_islands=CN.FalklandIslands,
    faroe_islands=CN.FaroeIslands,
    fiji=CN.Fiji,
    finland=CN.Finland,
    france=CN.France,
    french_guiana=CN.FrenchGuiana,
    french_polynesia=CN.FrenchPolynesia,
    french_southern_territories=CN.FrenchSouthernTerritories,
    gabon=CN.Gabon,
    gambia=CN.Gambia,
    georgia=CN.Georgia,
    germany=CN.Germany,
    ghana=CN.Ghana,
    gibraltar=CN.Gibraltar,
    greece=CN.Greece,
    greenland=CN.Greenland,
    grenada=CN.Grenada,
    guadeloupe=CN.Guadeloupe,
    guam=CN.Guam,
    guatemala=CN.Guatemala,
    guernsey=CN.Guernsey,
    guinea=CN.Guinea,
    guineabissau=CN.GuineaBissau,
    guyana=CN.Guyana,
    haiti=CN.Haiti,
    heard_and_mcdonald_islands=CN.HeardAndMcDonaldIslands,
    honduras=CN.Honduras,
    hongkong=CN.HongKong,
    hungary=CN.Hungary,
    iceland=CN.Iceland,
    india=CN.India,
    indonesia=CN.Indonesia,
    iran=CN.Iran,
    iraq=CN.Iraq,
    ireland=CN.Ireland,
    isle_of_man=CN.IsleOfMan,
    israel=CN.Israel,
    italy=CN.Italy,
    ivory_coast=CN.IvoryCoast,
    jamaica=CN.Jamaica,
    japan=CN.Japan,
    jersey=CN.Jersey,
    jordan=CN.Jordan,
    kazakhstan=CN.Kazakhstan,
    kenya=CN.Kenya,
    kiribati=CN.Kiribati,
    kosovo=CN.Kosovo,
    kuwait=CN.Kuwait,
    kyrgyzstan=CN.Kyrgyzstan,
    laos=CN.Laos,
    latin_america=CN.LatinAmerica,
    latvia=CN.Latvia,
    lebanon=CN.Lebanon,
    lesotho=CN.Lesotho,
    liberia=CN.Liberia,
    libya=CN.Libya,
    liechtenstein=CN.Liechtenstein,
    lithuania=CN.Lithuania,
    luxembourg=CN.Luxembourg,
    macau=CN.Macau,
    macedonia=CN.Macedonia,
    madagascar=CN.Madagascar,
    malawi=CN.Malawi,
    malaysia=CN.Malaysia,
    maldives=CN.Maldives,
    mali=CN.Mali,
    malta=CN.Malta,
    marshall_islands=CN.MarshallIslands,
    martinique=CN.Martinique,
    mauritania=CN.Mauritania,
    mauritius=CN.Mauritius,
    mayotte=CN.Mayotte,
    mexico=CN.Mexico,
    micronesia=CN.Micronesia,
    moldova=CN.Moldova,
    monaco=CN.Monaco,
    mongolia=CN.Mongolia,
    montenegro=CN.Montenegro,
    montserrat=CN.Montserrat,
    morocco=CN.Morocco,
    mozambique=CN.Mozambique,
    myanmar=CN.Myanmar,
    namibia=CN.Namibia,
    nauru_country=CN.NauruCountry,
    nepal=CN.Nepal,
    netherlands=CN.Netherlands,
    new_caledonia=CN.NewCaledonia,
    new_zealand=CN.NewZealand,
    nicaragua=CN.Nicaragua,
    niger=CN.Niger,
    nigeria=CN.Nigeria,
    niue=CN.Niue,
    norfolk_island=CN.NorfolkIsland,
    northern_mariana_islands=CN.NorthernMarianaIslands,
    north_korea=CN.NorthKorea,
    norway=CN.Norway,
    oman=CN.Oman,
    outlying_oceania=CN.OutlyingOceania,
    pakistan=CN.Pakistan,
    palau=CN.Palau,
    palestinian_territories=CN.PalestinianTerritories,
    panama=CN.Panama,
    papua_new_guinea=CN.PapuaNewGuinea,
    paraguay=CN.Paraguay,
    peru=CN.Peru,
    philippines=CN.Philippines,
    pitcairn=CN.Pitcairn,
    poland=CN.Poland,
    portugal=CN.Portugal,
    puertorico=CN.PuertoRico,
    qatar=CN.Qatar,
    reunion=CN.Reunion,
    romania=CN.Romania,
    russia=CN.Russia,
    rwanda=CN.Rwanda,
    saint_barthelemy=CN.SaintBarthelemy,
    saint_helena=CN.SaintHelena,
    saint_kitts_and_nevis=CN.SaintKittsAndNevis,
    saint_lucia=CN.SaintLucia,
    saint_martin=CN.SaintMartin,
    saint_pierre_and_miquelon=CN.SaintPierreAndMiquelon,
    saint_vincent_and_the_grenadines=CN.SaintVincentAndTheGrenadines,
    samoa=CN.Samoa,
    san_marino=CN.SanMarino,
    sao_tome_and_principe=CN.SaoTomeAndPrincipe,
    saudi_arabia=CN.SaudiArabia,
    senegal=CN.Senegal,
    serbia=CN.Serbia,
    seychelles=CN.Seychelles,
    sierra_leone=CN.SierraLeone,
    singapore=CN.Singapore,
    sint_maarten=CN.SintMaarten,
    slovakia=CN.Slovakia,
    slovenia=CN.Slovenia,
    solomon_islands=CN.SolomonIslands,
    somalia=CN.Somalia,
    south_africa=CN.SouthAfrica,
    south_georgia_and_south_sandwich_islands=CN.SouthGeorgiaAndTheSouthSandwichIslands,
    south_korea=CN.SouthKorea,
    south_sudan=CN.SouthSudan,
    spain=CN.Spain,
    sri_lanka=CN.SriLanka,
    sudan=CN.Sudan,
    suriname=CN.Suriname,
    svalbard_and_jan_mayen_islands=CN.SvalbardAndJanMayenIslands,
    swaziland=CN.Swaziland,
    sweden=CN.Sweden,
    switzerland=CN.Switzerland,
    syria=CN.Syria,
    taiwan=CN.Taiwan,
    tajikistan=CN.Tajikistan,
    tanzania=CN.Tanzania,
    thailand=CN.Thailand,
    togo=CN.Togo,
    tokelau_country=CN.TokelauCountry,
    tonga=CN.Tonga,
    trinidad_and_tobago=CN.TrinidadAndTobago,
    tristan_da_cunha=CN.TristanDaCunha,
    tunisia=CN.Tunisia,
    turkey=CN.Turkey,
    turkmenistan=CN.Turkmenistan,
    turks_and_caicos_islands=CN.TurksAndCaicosIslands,
    tuvalu_country=CN.TuvaluCountry,
    uganda=CN.Uganda,
    ukraine=CN.Ukraine,
    united_arabemirates=CN.UnitedArabEmirates,
    united_kingdom=CN.UnitedKingdom,
    united_states=CN.UnitedStates,
    united_states_minor_outlying_islands=CN.UnitedStatesMinorOutlyingIslands,
    united_states_virgin_islands=CN.UnitedStatesVirginIslands,
    uruguay=CN.Uruguay,
    uzbekistan=CN.Uzbekistan,
    vanuatu=CN.Vanuatu,
    vatican_city_state=CN.VaticanCityState,
    venezuela=CN.Venezuela,
    vietnam=CN.Vietnam,
    wallis_and_futuna_islands=CN.WallisAndFutunaIslands,
    western_sahara=CN.WesternSahara,
    world=CN.World,
    yemen=CN.Yemen,
    zambia=CN.Zambia,
    zimbabwe=CN.Zimbabwe,
)

LN = Loc.Language

LANGUAGE = bidict(
    any_language=LN.AnyLanguage,
    c=LN.C,
    abkhazian=LN.Abkhazian,
    afar=LN.Afar,
    afrikaans=LN.Afrikaans,
    aghem=LN.Aghem,
    akan=LN.Akan,
    akkadian=LN.Akkadian,
    akoose=LN.Akoose,
    albanian=LN.Albanian,
    american_sign_language=LN.AmericanSignLanguage,
    amharic=LN.Amharic,
    ancient_egyptian=LN.AncientEgyptian,
    ancient_greek=LN.AncientGreek,
    arabic=LN.Arabic,
    aragonese=LN.Aragonese,
    aramaic=LN.Aramaic,
    armenian=LN.Armenian,
    assamese=LN.Assamese,
    asturian=LN.Asturian,
    asu=LN.Asu,
    atsam=LN.Atsam,
    avaric=LN.Avaric,
    avestan=LN.Avestan,
    aymara=LN.Aymara,
    azerbaijani=LN.Azerbaijani,
    bafia=LN.Bafia,
    balinese=LN.Balinese,
    bambara=LN.Bambara,
    bamun=LN.Bamun,
    basaa=LN.Basaa,
    bashkir=LN.Bashkir,
    basque=LN.Basque,
    bataktoba=LN.BatakToba,
    belarusian=LN.Belarusian,
    bemba=LN.Bemba,
    bena=LN.Bena,
    bengali=LN.Bengali,
    bhojpuri=LN.Bhojpuri,
    bislama=LN.Bislama,
    blin=LN.Blin,
    bodo=LN.Bodo,
    bosnian=LN.Bosnian,
    breton=LN.Breton,
    buginese=LN.Buginese,
    bulgarian=LN.Bulgarian,
    burmese=LN.Burmese,
    cantonese=LN.Cantonese,
    catalan=LN.Catalan,
    cebuano=LN.Cebuano,
    central_kurdish=LN.CentralKurdish,
    central_morocco_tamazight=LN.CentralMoroccoTamazight,
    chakma=LN.Chakma,
    chamorro=LN.Chamorro,
    chechen=LN.Chechen,
    cherokee=LN.Cherokee,
    chickasaw=LN.Chickasaw,
    chiga=LN.Chiga,
    chinese=LN.Chinese,
    church=LN.Church,
    chuvash=LN.Chuvash,
    colognian=LN.Colognian,
    coptic=LN.Coptic,
    cornish=LN.Cornish,
    corsican=LN.Corsican,
    cree=LN.Cree,
    croatian=LN.Croatian,
    czech=LN.Czech,
    danish=LN.Danish,
    divehi=LN.Divehi,
    dogri=LN.Dogri,
    duala=LN.Duala,
    dutch=LN.Dutch,
    dzongkha=LN.Dzongkha,
    embu=LN.Embu,
    english=LN.English,
    erzya=LN.Erzya,
    esperanto=LN.Esperanto,
    estonian=LN.Estonian,
    ewe=LN.Ewe,
    ewondo=LN.Ewondo,
    faroese=LN.Faroese,
    fijian=LN.Fijian,
    filipino=LN.Filipino,
    finnish=LN.Finnish,
    french=LN.French,
    friulian=LN.Friulian,
    fulah=LN.Fulah,
    ga=LN.Ga,
    gaelic=LN.Gaelic,
    galician=LN.Galician,
    ganda=LN.Ganda,
    geez=LN.Geez,
    georgian=LN.Georgian,
    german=LN.German,
    gothic=LN.Gothic,
    greek=LN.Greek,
    greenlandic=LN.Greenlandic,
    guarani=LN.Guarani,
    gujarati=LN.Gujarati,
    gusii=LN.Gusii,
    haitian=LN.Haitian,
    hausa=LN.Hausa,
    hawaiian=LN.Hawaiian,
    hebrew=LN.Hebrew,
    herero=LN.Herero,
    hindi=LN.Hindi,
    hirimotu=LN.HiriMotu,
    hungarian=LN.Hungarian,
    icelandic=LN.Icelandic,
    ido=LN.Ido,
    igbo=LN.Igbo,
    inari_sami=LN.InariSami,
    indonesian=LN.Indonesian,
    ingush=LN.Ingush,
    interlingua=LN.Interlingua,
    interlingue=LN.Interlingue,
    inuktitut=LN.Inuktitut,
    inupiak=LN.Inupiak,
    irish=LN.Irish,
    italian=LN.Italian,
    japanese=LN.Japanese,
    javanese=LN.Javanese,
    jju=LN.Jju,
    jola_fonyi=LN.JolaFonyi,
    kabuverdianu=LN.Kabuverdianu,
    kabyle=LN.Kabyle,
    kako=LN.Kako,
    kalenjin=LN.Kalenjin,
    kamba=LN.Kamba,
    kannada=LN.Kannada,
    kanuri=LN.Kanuri,
    kashmiri=LN.Kashmiri,
    kazakh=LN.Kazakh,
    kenyang=LN.Kenyang,
    khmer=LN.Khmer,
    kiche=LN.Kiche,
    kikuyu=LN.Kikuyu,
    kinyarwanda=LN.Kinyarwanda,
    kirghiz=LN.Kirghiz,
    komi=LN.Komi,
    kongo=LN.Kongo,
    konkani=LN.Konkani,
    korean=LN.Korean,
    koro=LN.Koro,
    koyraboro_senni=LN.KoyraboroSenni,
    koyra_chiini=LN.KoyraChiini,
    kpelle=LN.Kpelle,
    kurdish=LN.Kurdish,
    kwanyama=LN.Kwanyama,
    kwasio=LN.Kwasio,
    lakota=LN.Lakota,
    langi=LN.Langi,
    lao=LN.Lao,
    latin=LN.Latin,
    latvian=LN.Latvian,
    lezghian=LN.Lezghian,
    limburgish=LN.Limburgish,
    lingala=LN.Lingala,
    literary_chinese=LN.LiteraryChinese,
    lithuanian=LN.Lithuanian,
    lojban=LN.Lojban,
    lower_sorbian=LN.LowerSorbian,
    low_german=LN.LowGerman,
    luba_katanga=LN.LubaKatanga,
    lule_sami=LN.LuleSami,
    luo=LN.Luo,
    luxembourgish=LN.Luxembourgish,
    luyia=LN.Luyia,
    macedonian=LN.Macedonian,
    machame=LN.Machame,
    maithili=LN.Maithili,
    makhuwa_meetto=LN.MakhuwaMeetto,
    makonde=LN.Makonde,
    malagasy=LN.Malagasy,
    malay=LN.Malay,
    malayalam=LN.Malayalam,
    maltese=LN.Maltese,
    mandingo=LN.Mandingo,
    manipuri=LN.Manipuri,
    manx=LN.Manx,
    maori=LN.Maori,
    mapuche=LN.Mapuche,
    marathi=LN.Marathi,
    marshallese=LN.Marshallese,
    masai=LN.Masai,
    mazanderani=LN.Mazanderani,
    mende=LN.Mende,
    meru=LN.Meru,
    meta=LN.Meta,
    mohawk=LN.Mohawk,
    mongolian=LN.Mongolian,
    morisyen=LN.Morisyen,
    mundang=LN.Mundang,
    muscogee=LN.Muscogee,
    nama=LN.Nama,
    nauru_language=LN.NauruLanguage,
    navaho=LN.Navaho,
    ndonga=LN.Ndonga,
    nepali=LN.Nepali,
    newari=LN.Newari,
    ngiemboon=LN.Ngiemboon,
    ngomba=LN.Ngomba,
    nko=LN.Nko,
    northern_luri=LN.NorthernLuri,
    northern_sami=LN.NorthernSami,
    northern_sotho=LN.NorthernSotho,
    north_ndebele=LN.NorthNdebele,
    norwegian_bokmal=LN.NorwegianBokmal,
    norwegian_nynorsk=LN.NorwegianNynorsk,
    nuer=LN.Nuer,
    nyanja=LN.Nyanja,
    nyankole=LN.Nyankole,
    occitan=LN.Occitan,
    ojibwa=LN.Ojibwa,
    old_irish=LN.OldIrish,
    old_norse=LN.OldNorse,
    old_persian=LN.OldPersian,
    oriya=LN.Oriya,
    oromo=LN.Oromo,
    osage=LN.Osage,
    ossetic=LN.Ossetic,
    pahlavi=LN.Pahlavi,
    palauan=LN.Palauan,
    pali=LN.Pali,
    papiamento=LN.Papiamento,
    pashto=LN.Pashto,
    persian=LN.Persian,
    phoenician=LN.Phoenician,
    polish=LN.Polish,
    portuguese=LN.Portuguese,
    prussian=LN.Prussian,
    punjabi=LN.Punjabi,
    quechua=LN.Quechua,
    romanian=LN.Romanian,
    romansh=LN.Romansh,
    rombo=LN.Rombo,
    rundi=LN.Rundi,
    russian=LN.Russian,
    rwa=LN.Rwa,
    saho=LN.Saho,
    sakha=LN.Sakha,
    samburu=LN.Samburu,
    samoan=LN.Samoan,
    sango=LN.Sango,
    sangu=LN.Sangu,
    sanskrit=LN.Sanskrit,
    santali=LN.Santali,
    sardinian=LN.Sardinian,
    saurashtra=LN.Saurashtra,
    sena=LN.Sena,
    serbian=LN.Serbian,
    shambala=LN.Shambala,
    shona=LN.Shona,
    sichuanyi=LN.SichuanYi,
    sicilian=LN.Sicilian,
    sidamo=LN.Sidamo,
    silesian=LN.Silesian,
    sindhi=LN.Sindhi,
    sinhala=LN.Sinhala,
    skoltsami=LN.SkoltSami,
    slovak=LN.Slovak,
    slovenian=LN.Slovenian,
    soga=LN.Soga,
    somali=LN.Somali,
    southern_kurdish=LN.SouthernKurdish,
    southern_sami=LN.SouthernSami,
    southern_sotho=LN.SouthernSotho,
    south_ndebele=LN.SouthNdebele,
    spanish=LN.Spanish,
    standard_moroccan_tamazight=LN.StandardMoroccanTamazight,
    sundanese=LN.Sundanese,
    swahili=LN.Swahili,
    swati=LN.Swati,
    swedish=LN.Swedish,
    swiss_german=LN.SwissGerman,
    syriac=LN.Syriac,
    tachelhit=LN.Tachelhit,
    tahitian=LN.Tahitian,
    taidam=LN.TaiDam,
    taita=LN.Taita,
    tajik=LN.Tajik,
    tamil=LN.Tamil,
    taroko=LN.Taroko,
    tasawaq=LN.Tasawaq,
    tatar=LN.Tatar,
    telugu=LN.Telugu,
    teso=LN.Teso,
    thai=LN.Thai,
    tibetan=LN.Tibetan,
    tigre=LN.Tigre,
    tigrinya=LN.Tigrinya,
    tokelau_language=LN.TokelauLanguage,
    tokpisin=LN.TokPisin,
    tongan=LN.Tongan,
    tsonga=LN.Tsonga,
    tswana=LN.Tswana,
    turkish=LN.Turkish,
    turkmen=LN.Turkmen,
    tuvalu_language=LN.TuvaluLanguage,
    tyap=LN.Tyap,
    ugaritic=LN.Ugaritic,
    uighur=LN.Uighur,
    ukrainian=LN.Ukrainian,
    upper_sorbian=LN.UpperSorbian,
    urdu=LN.Urdu,
    uzbek=LN.Uzbek,
    vai=LN.Vai,
    venda=LN.Venda,
    vietnamese=LN.Vietnamese,
    volapuk=LN.Volapuk,
    vunjo=LN.Vunjo,
    walamo=LN.Walamo,
    walloon=LN.Walloon,
    walser=LN.Walser,
    warlpiri=LN.Warlpiri,
    welsh=LN.Welsh,
    western_balochi=LN.WesternBalochi,
    western_frisian=LN.WesternFrisian,
    wolof=LN.Wolof,
    xhosa=LN.Xhosa,
    yangben=LN.Yangben,
    yiddish=LN.Yiddish,
    yoruba=LN.Yoruba,
    zarma=LN.Zarma,
    zhuang=LN.Zhuang,
    zulu=LN.Zulu,
)

FormatTypeStr = Literal["long", "short", "narrow"]

FORMAT_TYPE: bidict[FormatTypeStr, QtCore.QLocale.FormatType] = bidict(
    long=QtCore.QLocale.FormatType.LongFormat,
    short=QtCore.QLocale.FormatType.ShortFormat,
    narrow=QtCore.QLocale.FormatType.NarrowFormat,
)

MeasurementSystemStr = Literal["metric", "imperial_us", "imperial_uk"]

MEASUREMENT_SYSTEM: bidict[MeasurementSystemStr, QtCore.QLocale.MeasurementSystem] = (
    bidict(
        metric=QtCore.QLocale.MeasurementSystem.MetricSystem,
        imperial_us=QtCore.QLocale.MeasurementSystem.ImperialUSSystem,
        imperial_uk=QtCore.QLocale.MeasurementSystem.ImperialUKSystem,
    )
)

DataSizeFormatStr = Literal["iec", "traditional", "si"]

DATA_SIZE_FORMAT: bidict[DataSizeFormatStr, QtCore.QLocale.DataSizeIecFormat] = bidict(
    iec=QtCore.QLocale.DataSizeFormat.DataSizeIecFormat,
    traditional=QtCore.QLocale.DataSizeFormat.DataSizeTraditionalFormat,
    si=QtCore.QLocale.DataSizeFormat.DataSizeSIFormat,
)

NumberOptionStr = Literal[
    "default",
    "omit_group_separator",
    "reject_group_separator",
    "omit_leading_zero_in_exponent",
    "reject_leading_zero_in_exponent",
    "include_trailing_zeroes_after_dot",
    "reject_trailing_zeroes_after_dot",
]

NO = QtCore.QLocale.NumberOption
NUMBER_OPTION: bidict[NumberOptionStr, NO] = bidict(
    default=NO.DefaultNumberOptions,
    omit_group_separator=NO.OmitGroupSeparator,
    reject_group_separator=NO.RejectGroupSeparator,
    omit_leading_zero_in_exponent=NO.OmitLeadingZeroInExponent,
    reject_leading_zero_in_exponent=NO.RejectLeadingZeroInExponent,
    include_trailing_zeroes_after_dot=NO.IncludeTrailingZeroesAfterDot,
    reject_trailing_zeroes_after_dot=NO.RejectTrailingZeroesAfterDot,
)

OFFSET = ord("🇦") - ord("A")


class Locale(QtCore.QLocale):
    """Converts between numbers and their string representations in various languages."""

    def __repr__(self):
        return get_repr(self, self.bcp47Name())

    def __reduce__(self):
        return type(self), (self.bcp47Name(),)

    def get_flag_unicode(self):
        name = self.name().split("_")[1]
        name = [c for c in name.lower() if c.isalnum()]
        return "".join([chr(ord(c.upper()) + OFFSET) for c in name])

    @classmethod
    def get_system_locale(cls) -> Self:
        return cls(cls.system())

    @classmethod
    def get_c_locale(cls) -> Self:
        return cls(cls.c())

    @classmethod
    def get_system_language(cls) -> str:
        return cls.get_system_locale().uiLanguages()[0]

    def get_country(self):
        return COUNTRY.inverse[self.country()]

    def get_measurement_system(self) -> MeasurementSystemStr:
        return MEASUREMENT_SYSTEM.inverse[self.measurementSystem()]

    def get_formatted_data_size(
        self, size: int, precision: int = 2, fmt: DataSizeFormatStr = "iec"
    ) -> str:
        if size < 0:
            return ""
        return self.formattedDataSize(size, precision, DATA_SIZE_FORMAT[fmt])

    def get_first_day_of_week(self) -> constants.DayOfWeekStr:
        return constants.DAY_OF_WEEK.inverse[self.firstDayOfWeek()]

    def get_text_direction(self) -> constants.LayoutDirectionStr:
        return constants.LAYOUT_DIRECTION.inverse[self.textDirection()]

    def get_weekdays(self) -> list[constants.DayOfWeekStr]:
        return [constants.DAY_OF_WEEK.inverse[i] for i in self.weekdays()]

    def get_day_name(self, day: int, format_type: FormatTypeStr = "long") -> str:
        return self.dayName(day, FORMAT_TYPE[format_type])

    def get_month_name(self, month: int, format_type: FormatTypeStr = "long") -> str:
        return self.monthName(month, FORMAT_TYPE[format_type])

    def get_standalone_day_name(
        self, day: int, format_type: FormatTypeStr = "long"
    ) -> str:
        return self.standaloneDayName(day, FORMAT_TYPE[format_type])

    def get_time_format(self, format_type: FormatTypeStr = "long") -> str:
        return self.timeFormat(FORMAT_TYPE[format_type])

    def get_date_format(self, format_type: FormatTypeStr = "long") -> str:
        return self.dateFormat(FORMAT_TYPE[format_type])

    def get_datetime_format(self, format_type: FormatTypeStr = "long") -> str:
        return self.dateTimeFormat(FORMAT_TYPE[format_type])

    def to_datetime(
        self, text: str, format_type: FormatTypeStr | str = "long"
    ) -> QtCore.QDateTime:
        if format_type in {"long", "short", "narrow"}:
            format_type = FORMAT_TYPE[format_type]
        return self.toDateTime(text, format_type)

    def to_date(
        self, text: str, format_type: FormatTypeStr | str = "long"
    ) -> QtCore.QDate:
        if format_type in {"long", "short", "narrow"}:
            format_type = FORMAT_TYPE[format_type]
        return self.toDate(text, format_type)

    def to_time(
        self, text: str, format_type: FormatTypeStr | str = "long"
    ) -> QtCore.QTime:
        if format_type in {"long", "short", "narrow"}:
            format_type = FORMAT_TYPE[format_type]
        return self.to_time(text, format_type)

    def get_standalone_month_name(
        self, month: int, format_type: FormatTypeStr = "long"
    ) -> str:
        return self.standaloneMonthName(month, FORMAT_TYPE[format_type])

    def quote_string(self, string: str, alternate_style: bool = False) -> str:
        flag = (
            self.QuotationStyle.AlternateQuotation
            if alternate_style
            else self.QuotationStyle.StandardQuotation
        )
        return self.quoteString(string, flag)

    @classmethod
    def get_all_locales(cls) -> list[Self]:
        return [
            cls(i)
            for i in cls.matchingLocales(
                cls.Language.AnyLanguage, cls.Script.AnyScript, cls.Country.AnyCountry
            )
        ]


if __name__ == "__main__":
    from prettyqt import gui, widgets

    app = widgets.app()
    locale = Locale.get_all_locales()[100].get_flag_unicode()
    p = gui.Pixmap.create_char(locale, 64)
    w = widgets.Label()
    w.set_pixmap(p)
    w.show()
    app.exec()



from __future__ import annotations

from prettyqt import core, widgets


class DoubleSpinBox(widgets.AbstractSpinBoxMixin, widgets.QDoubleSpinBox):
    """Spin box widget that takes doubles."""

    value_changed = core.Signal(float)

    def __init__(self, *args, maximum: float = float("inf"), **kwargs):
        super().__init__(*args, maximum=maximum, **kwargs)
        self.valueChanged.connect(self.value_changed)

    @classmethod
    def supports(cls, instance) -> bool:
        return isinstance(instance, float)

    def set_range(self, start: float | None, end: float | None):
        self.set_minimum(start)
        self.set_maximum(end)

    setRange = set_range

    def set_minimum(self, value: float | None):
        if value is None:
            value = -float("inf")
        super().setMinimum(value)

    setMinimum = set_minimum

    def set_maximum(self, value: float | None):
        if value is None:
            value = float("inf")
        super().setMaximum(value)

    setMaximum = set_maximum


if __name__ == "__main__":
    app = widgets.app()
    widget = DoubleSpinBox()
    widget.show()
    app.exec()



from __future__ import annotations

from prettyqt import core, widgets


class StringOrNumberWidget(widgets.GroupBox):
    value_changed = core.Signal(object)

    def __init__(self, *args, object_name: str = "string_or_number_widget", **kwargs):
        super().__init__(*args, object_name=object_name, **kwargs)
        self.set_layout("vertical")
        self.rb_lineedit = widgets.RadioButton("String", checked=True)
        self.lineedit = widgets.LineEdit()
        self.rb_spinbox = widgets.RadioButton("Number")
        self.spinbox = widgets.DoubleSpinBox()
        layout_lineedit = widgets.HBoxLayout()
        layout_lineedit.add(self.rb_lineedit)
        layout_lineedit.add(self.lineedit)
        layout_spinbox = widgets.HBoxLayout()
        layout_spinbox.add(self.rb_spinbox)
        layout_spinbox.add(self.spinbox)
        self.box.add(layout_lineedit)
        self.box.add(layout_spinbox)
        self.rb_spinbox.toggled.connect(self.spinbox.setEnabled)
        self.rb_spinbox.toggled.connect(self.lineedit.setDisabled)
        self.rb_lineedit.toggled.connect(self.lineedit.setEnabled)
        self.rb_lineedit.toggled.connect(self.spinbox.setDisabled)
        self.spinbox.value_changed.connect(self.on_value_change)
        self.lineedit.value_changed.connect(self.on_value_change)

    def on_value_change(self):
        value = self.get_value()
        self.value_changed.emit(value)

    def get_value(self) -> float | str:
        if not self.rb_spinbox.isChecked():
            return self.lineedit.get_value()
        val = self.spinbox.get_value()
        return int(val) if val.is_integer() else val

    def set_value(self, value: float | str):
        match value:
            case str():
                self.rb_lineedit.setChecked(True)
                self.lineedit.set_value(value)
            case int() | float():
                self.rb_spinbox.setChecked(True)
                self.spinbox.set_value(value)
            case _:
                msg = f"Invalid Type for set_value: {type(value)}"
                raise TypeError(msg)


if __name__ == "__main__":
    app = widgets.app()
    widget = StringOrNumberWidget("Test")
    widget.value_changed.connect(print)
    widget.show()
    app.exec()



from __future__ import annotations

import os
import pathlib
from typing import TYPE_CHECKING, Literal

from prettyqt import core
from prettyqt.utils import bidict, datatypes, get_repr


if TYPE_CHECKING:
    from collections.abc import Iterator


CompressionStr = Literal["none", "zlib", "zstd"]

COMPRESSION: bidict[CompressionStr, core.QResource.Compression] = bidict(
    none=core.QResource.Compression.NoCompression,
    zlib=core.QResource.Compression.ZlibCompression,
    zstd=core.QResource.Compression.ZstdCompression,
)


class Resource(core.QResource):
    """Interface for reading directly from resources."""

    def __repr__(self):
        return get_repr(self, self.absoluteFilePath(), self.get_locale())

    def __reduce__(self):
        return type(self), (self.absoluteFilePath(), self.get_locale())

    def __bytes__(self):
        return bytes(self.uncompressedData())

    def __bool__(self):
        return self.isValid()

    def __iter__(self) -> Iterator[str]:
        return iter(self.children())

    def get_compression_algorithm(self) -> CompressionStr:
        return COMPRESSION.inverse[self.compressionAlgorithm()]

    def get_absolute_file_path(self) -> pathlib.Path:
        return pathlib.Path(self.absoluteFilePath())

    def get_locale(self) -> core.Locale:
        return core.Locale(self.locale())

    def get_last_modified(self) -> core.DateTime:
        return core.DateTime(self.lastModified())

    def set_file_name(self, path: datatypes.PathType):
        self.setFileName(os.fspath(path))

    def get_file_name(self) -> pathlib.Path:
        return pathlib.Path(self.fileName())

    @classmethod
    def register_resource(cls, path: datatypes.PathType, root: str = ""):
        cls.registerResource(os.fspath(path), root)



from __future__ import annotations

from prettyqt import gui, widgets


class StylePainter(gui.PainterMixin, widgets.QStylePainter):
    """Convenience class for drawing QStyle elements inside a widget."""

    def draw_complex_control(
        self,
        control: widgets.style.ComplexControlStr,
        option: widgets.QStyleOptionComplex,
    ):
        self.drawComplexControl(widgets.style.COMPLEX_CONTROL[control], option)



from __future__ import annotations

from prettyqt.qt import QtCore
from prettyqt.utils import get_repr


class ByteArrayMatcher(QtCore.QByteArrayMatcher):
    """Holds a sequence of bytes that can be quickly matched in a byte array."""

    def __repr__(self):
        return get_repr(self, self.get_pattern())

    def get_pattern(self) -> bytes:
        return bytes(self.pattern())


if __name__ == "__main__":
    matcher = ByteArrayMatcher(QtCore.QByteArray(b"Test"))



from __future__ import annotations

from typing import Literal

from prettyqt import core, network
from prettyqt.utils import bidict


DnsEligibilityStatusStr = Literal["unknown", "eligible", "ineligible"]

DNS_ELIGIBILITY_STATUS: bidict[
    DnsEligibilityStatusStr, network.QNetworkAddressEntry.DnsEligibilityStatus
] = bidict(
    unknown=network.QNetworkAddressEntry.DnsEligibilityStatus.DnsEligibilityUnknown,
    eligible=network.QNetworkAddressEntry.DnsEligibilityStatus.DnsEligible,
    ineligible=network.QNetworkAddressEntry.DnsEligibilityStatus.DnsIneligible,
)


class NetworkAddressEntry(network.QNetworkAddressEntry):
    """Stores one IP address supported by a network interface."""

    def get_dns_eligibility(self) -> DnsEligibilityStatusStr:
        """Return whether this address is eligible for publication in the DNS.

        Returns:
            DNS eligibility
        """
        return DNS_ELIGIBILITY_STATUS.inverse[self.dnsEligibility()]

    def set_dns_eligibility(
        self,
        status: DnsEligibilityStatusStr
        | network.QNetworkAddressEntry.DnsEligibilityStatus,
    ):
        """Set the DNS eligibility flag for this address to status.

        Args:
            status: DNS eligibility status
        """
        self.setDnsEligibility(DNS_ELIGIBILITY_STATUS.get_enum_value(status))

    def get_ip(self) -> network.HostAddress:
        return network.HostAddress(self.ip())

    def set_ip(self, ip: network.QHostAddress | str):
        self.setIp(network.HostAddress(ip))

    def get_netmask(self) -> network.HostAddress:
        return network.HostAddress(self.netmask())

    def set_netmask(self, netmask: network.QHostAddress | str):
        self.setNetmask(network.HostAddress(netmask))

    def get_preferred_lifetime(self) -> core.DeadlineTimer:
        return core.DeadlineTimer(self.preferredLifetime())

    def get_validity_lifetime(self) -> core.DeadlineTimer:
        return core.DeadlineTimer(self.validityLifetime())



from __future__ import annotations

from typing import TYPE_CHECKING

from prettyqt import gui, iconprovider, widgets
from prettyqt.utils import datatypes, listdelegators


if TYPE_CHECKING:
    from collections.abc import Iterator


class ToolBox(widgets.FrameMixin, widgets.QToolBox):
    """Column of tabbed widget items."""

    def __getitem__(
        self, index: int | str
    ) -> widgets.QWidget | listdelegators.ListDelegator[widgets.QWidget]:
        match index:
            case int():
                return self.widget(index)
            case str():
                result = self.find_child(widgets.QWidget, index)
                if result is None:
                    msg = "Widget not found"
                    raise KeyError(msg)
                return result
            case slice():
                stop = index.stop or self.count()
                rng = range(index.start or 0, stop, index.step or 1)
                wdgs = [self.widget(i) for i in rng]
                return listdelegators.ListDelegator(wdgs)
            case _:
                raise TypeError(index)

    def __delitem__(self, index: int):
        self.removeItem(index)

    # def __setstate__(self, state):
    #     for i, item in enumerate(state["items"]):
    #         self.addItem(item["widget"], item["icon"], item["text"])
    #         self.setItemEnabled(i, item["enabled"])
    #         self.setItemToolTip(i, item["tool_tip"])
    #     self.setCurrentIndex(state["current_index"])

    # def __reduce__(self):
    #     return type(self), (), self.__getstate__()

    def __iter__(self) -> Iterator[widgets.QWidget]:
        return iter(self.get_children())

    def __contains__(self, item: widgets.QWidget):
        return self.indexOf(item) >= 0

    def get_children(self) -> listdelegators.ListDelegator[widgets.QWidget]:
        widgets = [self.widget(i) for i in range(self.count())]
        return listdelegators.ListDelegator(widgets)

    def add_widget(
        self,
        widget: widgets.QWidget,
        title: str | None = None,
        icon: datatypes.IconType = None,
        tooltip: str = "",
        enabled: bool = True,
    ):
        title = widget.objectName() if title is None else title
        icon = iconprovider.get_icon(icon) if icon else gui.Icon()
        self.addItem(widget, icon, title)
        index = self.indexOf(widget)
        if tooltip:
            self.setItemToolTip(index, tooltip)
        self.setItemEnabled(index, enabled)

    def get_item_icon(self, index: int) -> gui.Icon | None:
        icon = self.itemIcon(index)
        return None if icon.isNull() else gui.Icon(icon)


if __name__ == "__main__":
    app = widgets.app()
    w = widgets.Widget()
    w2 = widgets.Widget()
    tb = ToolBox()
    tb.add_widget(w, "title")
    tb.add_widget(w2)
    tb.show()
    app.exec()



"""Base classes for graphical user interface (GUI) components."""

from __future__ import annotations

import sys

from prettyqt.qt.QtGui import *  # noqa: F403

from .backingstore import BackingStore
from .stylehints import StyleHints
from .pageranges import PageRanges
from .shortcut import Shortcut
from .textobjectinterface import TextObjectInterface
from .drag import Drag
from .screen import Screen
from .keysequence import KeySequence
from .surfaceformat import SurfaceFormat
from .surface import Surface, SurfaceMixin
from .window import Window, WindowMixin
from .icon import Icon
from .paintdevice import PaintDevice, PaintDeviceMixin
from .paintdevicewindow import PaintDeviceWindow, PaintDeviceWindowMixin
from .pixmap import Pixmap, PixmapMixin
from .iconengine import IconEngine
from .pixmapcache import PixmapCache
from .bitmap import Bitmap
from .image import Image

from .rasterwindow import RasterWindow
from .clipboard import Clipboard
from .inputmethod import InputMethod
from .sessionmanager import SessionManager
from .validator import Validator, ValidatorMixin

from .regularexpressionvalidator import RegularExpressionValidator
from .textlayout import TextLayout
from .textline import TextLine
from .textoption import TextOption
from .textblock import TextBlock
from .textdocumentwriter import TextDocumentWriter
from .textdocument import TextDocument
from .textdocumentfragment import TextDocumentFragment
from .statictext import StaticText
from .intvalidator import IntValidator
from .doublevalidator import DoubleValidator
from .color import Color
from .brush import Brush
from .textblockuserdata import TextBlockUserData
from .gradient import Gradient, GradientMixin
from .lineargradient import LinearGradient
from .radialgradient import RadialGradient
from .conicalgradient import ConicalGradient
from .pagesize import PageSize
from .pagelayout import PageLayout
from .font import Font
from .fontmetrics import FontMetrics
from .fontmetricsf import FontMetricsF
from .fontinfo import FontInfo
from .fontdatabase import FontDatabase
from .region import Region
from .movie import Movie
from .transform import Transform
from .pagedpaintdevice import PagedPaintDevice, PagedPaintDeviceMixin
from .pen import Pen
from .picture import Picture
from .polygon import Polygon
from .polygonf import PolygonF
from .painter import Painter, PainterMixin
from .painterpath import PainterPath
from .painterpathstroker import PainterPathStroker
from .palette import Palette
from .guiapplication import GuiApplication, GuiApplicationMixin
from .cursor import Cursor
from .standarditem import StandardItem
from .standarditemmodel import StandardItemModel
from .textcursor import TextCursor
from .pdfwriter import PdfWriter
from .desktopservices import DesktopServices
from .matrix4x4 import Matrix4x4
from .vector3d import Vector3D
from .vector4d import Vector4D
from .imageiohandler import ImageIOHandler
from .imagereader import ImageReader
from .imagewriter import ImageWriter
from .textformat import TextFormat, TextFormatMixin
from .textlength import TextLength
from .textframeformat import TextFrameFormat
from .textblockformat import TextBlockFormat
from .textcharformat import TextCharFormat, TextCharFormatMixin
from .textimageformat import TextImageFormat
from .textlistformat import TextListFormat
from .texttablecellformat import TextTableCellFormat
from .textobject import TextObject, TextObjectMixin
from .textblockgroup import TextBlockGroup
from .textframe import TextFrame
from .texttablecell import TextTableCell
from .texttableformat import TextTableFormat
from .texttable import TextTable
from .abstracttextdocumentlayout import (
    AbstractTextDocumentLayout,
    AbstractTextDocumentLayoutMixin,
)
from .abstractfileiconprovider import (
    AbstractFileIconProvider,
    AbstractFileIconProviderMixin,
)
from .syntaxhighlighter import SyntaxHighlighter
from .undocommand import UndoCommand
from .undostack import UndoStack
from .undogroup import UndoGroup
from .colorspace import ColorSpace
from .action import Action, ActionMixin
from .actiongroup import ActionGroup
from prettyqt.qt import QtGui

QT_MODULE = QtGui
# might_be_rich_text = Qt.mightBeRichText


def app(args: list[str] | None = None, **kwargs) -> GuiApplication:
    if (instance := GuiApplication.instance()) is not None:
        return instance
    return GuiApplication(sys.argv if args is None else args, **kwargs)


__all__ = [
    "AbstractFileIconProvider",
    "AbstractFileIconProviderMixin",
    "AbstractTextDocumentLayout",
    "AbstractTextDocumentLayoutMixin",
    "Action",
    "ActionGroup",
    "ActionMixin",
    "BackingStore",
    "Bitmap",
    "Brush",
    "Clipboard",
    "Color",
    "ColorSpace",
    "ConicalGradient",
    "Cursor",
    "DesktopServices",
    "DoubleValidator",
    "Drag",
    "Font",
    "FontDatabase",
    "FontInfo",
    "FontMetrics",
    "FontMetricsF",
    "Gradient",
    "GradientMixin",
    "GuiApplication",
    "GuiApplicationMixin",
    "Icon",
    "IconEngine",
    "Image",
    "ImageIOHandler",
    "ImageReader",
    "ImageWriter",
    "InputMethod",
    "IntValidator",
    "KeySequence",
    "LinearGradient",
    "Matrix4x4",
    "Movie",
    "PageLayout",
    "PageRanges",
    "PageSize",
    "PagedPaintDevice",
    "PagedPaintDeviceMixin",
    "PaintDevice",
    "PaintDeviceMixin",
    "PaintDeviceWindow",
    "PaintDeviceWindowMixin",
    "Painter",
    "PainterMixin",
    "PainterPath",
    "PainterPathStroker",
    "Palette",
    "PdfWriter",
    "Pen",
    "Picture",
    "Pixmap",
    "PixmapCache",
    "PixmapMixin",
    "Polygon",
    "PolygonF",
    "RadialGradient",
    "RasterWindow",
    "Region",
    "RegularExpressionValidator",
    "Screen",
    "SessionManager",
    "Shortcut",
    "StandardItem",
    "StandardItemModel",
    "StaticText",
    "StyleHints",
    "Surface",
    "SurfaceFormat",
    "SurfaceMixin",
    "SyntaxHighlighter",
    "TextBlock",
    "TextBlockFormat",
    "TextBlockGroup",
    "TextBlockUserData",
    "TextCharFormat",
    "TextCharFormatMixin",
    "TextCursor",
    "TextDocument",
    "TextDocumentFragment",
    "TextDocumentWriter",
    "TextFormat",
    "TextFormatMixin",
    "TextFrame",
    "TextFrameFormat",
    "TextImageFormat",
    "TextLayout",
    "TextLength",
    "TextLine",
    "TextListFormat",
    "TextObject",
    "TextObjectInterface",
    "TextObjectMixin",
    "TextOption",
    "TextTable",
    "TextTableCell",
    "TextTableCellFormat",
    "TextTableFormat",
    "Transform",
    "UndoCommand",
    "UndoGroup",
    "UndoStack",
    "Validator",
    "ValidatorMixin",
    "Vector3D",
    "Vector4D",
    "Window",
    "WindowMixin",
    "app",
]



from __future__ import annotations

from collections.abc import Sequence

from prettyqt import constants, core, itemmodels


class RootPathColumn(itemmodels.ColumnItem):
    name = "Root path"
    doc = "Root path"

    def get_data(self, item: core.QStorageInfo, role: constants.ItemDataRole):
        match role:
            case constants.DISPLAY_ROLE:
                return item.rootPath()


class VolumeNameColumn(itemmodels.ColumnItem):
    name = "Volume name"
    doc = "Volume name"

    def get_data(self, item: core.QStorageInfo, role: constants.ItemDataRole):
        match role:
            case constants.DISPLAY_ROLE:
                return item.name()


class DeviceColumn(itemmodels.ColumnItem):
    name = "Device"

    def get_data(self, item: core.QStorageInfo, role: constants.ItemDataRole):
        match role:
            case constants.DISPLAY_ROLE:
                return item.device().data().decode()


class FileSystemColumn(itemmodels.ColumnItem):
    name = "File system"

    def get_data(self, item: core.QStorageInfo, role: constants.ItemDataRole):
        match role:
            case constants.DISPLAY_ROLE:
                return item.fileSystemType().data().decode()


class TotalColumn(itemmodels.ColumnItem):
    name = "Total"
    precision = 2
    fmt: core._locale.DataSizeFormatStr = "iec"

    def get_data(self, item: core.QStorageInfo, role: constants.ItemDataRole):
        match role:
            case constants.DISPLAY_ROLE:
                return core.Locale().get_formatted_data_size(
                    item.bytesTotal(), precision=self.precision, fmt=self.fmt
                )


class FreeColumn(itemmodels.ColumnItem):
    name = "Free"
    precision = 2
    fmt: core._locale.DataSizeFormatStr = "iec"

    def get_data(self, item: core.QStorageInfo, role: constants.ItemDataRole):
        match role:
            case constants.DISPLAY_ROLE:
                return core.Locale().get_formatted_data_size(
                    item.bytesFree(), precision=self.precision, fmt=self.fmt
                )


class AvailableColumn(itemmodels.ColumnItem):
    name = "Available"
    precision = 2
    fmt: core._locale.DataSizeFormatStr = "iec"

    def get_data(self, item: core.QStorageInfo, role: constants.ItemDataRole):
        match role:
            case constants.DISPLAY_ROLE:
                return core.Locale().get_formatted_data_size(
                    item.bytesAvailable(), precision=self.precision, fmt=self.fmt
                )


class ReadyColumn(itemmodels.ColumnItem):
    name = "Is ready"

    def get_data(self, item: core.QStorageInfo, role: constants.ItemDataRole):
        match role:
            case constants.CHECKSTATE_ROLE:
                return self.to_checkstate(item.isReady())


class ReadOnlyColumn(itemmodels.ColumnItem):
    name = "Read-only"

    def get_data(self, item: core.QStorageInfo, role: constants.ItemDataRole):
        match role:
            case constants.CHECKSTATE_ROLE:
                return self.to_checkstate(item.isReadOnly())


class ValidColumn(itemmodels.ColumnItem):
    name = "Valid"

    def get_data(self, item: core.QStorageInfo, role: constants.ItemDataRole):
        match role:
            case constants.CHECKSTATE_ROLE:
                return self.to_checkstate(item.isValid())


class StorageInfoModel(itemmodels.ColumnTableModel):
    """Model to display a list of core.QStorageInfos."""

    COLUMNS = [
        RootPathColumn,
        VolumeNameColumn,
        DeviceColumn,
        FileSystemColumn,
        TotalColumn,
        FreeColumn,
        AvailableColumn,
        ReadyColumn,
        ReadOnlyColumn,
        ValidColumn,
    ]
    SUPPORTS = Sequence[core.QStorageInfo]

    def __init__(self, volumes: list[core.QStorageInfo], parent=None):
        super().__init__(volumes, self.COLUMNS, parent=parent)

    @classmethod
    def supports(cls, instance) -> bool:
        match instance:
            case (core.QStorageInfo(), *_):
                return True
            case _:
                return False


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    view = widgets.TreeView()
    volumes = core.StorageInfo.get_mounted_volumes()
    model = StorageInfoModel(volumes, parent=view)
    view.setModel(model)
    view.resize(640, 480)
    view.set_selection_behavior("rows")
    view.adapt_sizes()
    view.show()
    app.exec()



from __future__ import annotations

import os
import pathlib
from typing import TYPE_CHECKING, Literal

from prettyqt import constants, core, qt, widgets
from prettyqt.utils import bidict, datatypes


if TYPE_CHECKING:
    from collections.abc import Sequence


OptionStr = Literal["dont_watch_changes", "dont_resolve_symlinks", "no_custom_icons"]

OPTIONS: bidict[OptionStr, widgets.QFileSystemModel.Option] = bidict(
    dont_watch_changes=widgets.QFileSystemModel.Option.DontWatchForChanges,
    dont_resolve_symlinks=widgets.QFileSystemModel.Option.DontResolveSymlinks,
    no_custom_icons=widgets.QFileSystemModel.Option.DontUseCustomDirectoryIcons,
)


class FileSystemModelMixin:
    """Class to populate a filesystem treeview."""

    DATA_ROLE = constants.USER_ROLE + 33  # type: ignore
    content_type = "files"

    def __init__(self, *args, read_only: bool = False, **kwargs):
        super().__init__(*args, read_only=read_only, **kwargs)
        self.use_custom_icons(False)

    def data(
        self,
        index: core.ModelIndex,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        if role == constants.USER_ROLE:
            path = index.data(self.Roles.FilePathRole)
            return pathlib.Path(path)
        return super().data(index, role)

    def parent(self, *args):
        # workaround: PyQt6 QFileSystemModel.parent() missing
        if not args and qt.API == "pyqt6":
            return core.QAbstractItemModel.parent(self)
        return super().parent(*args)

    def get_file_info(self, index: core.ModelIndex) -> core.FileInfo:
        return core.FileInfo(self.fileInfo(index))

    def get_file_path(self, index: core.ModelIndex) -> pathlib.Path:
        return pathlib.Path(self.filePath(index))

    def resolve_sym_links(self, resolve: bool):
        self.setResolveSymlinks(resolve)

    def watch_for_changes(self, watch: bool):
        self.setOption(OPTIONS["dont_watch_changes"], not watch)

    def use_custom_icons(self, use: bool):
        self.setOption(OPTIONS["no_custom_icons"], not use)

    def set_root_path(self, path: datatypes.PathType) -> core.ModelIndex:
        match path:
            case "root":
                path = core.Dir.rootPath()
            case "home":
                path = core.Dir.homePath()
            case "temp":
                path = core.Dir.tempPath()
            case "current":
                path = core.Dir.currentPath()
            case _:
                path = os.fspath(path)
        return self.setRootPath(path)

    def set_name_filters(self, filters, hide: bool = False):
        self.setNameFilters(filters)
        self.setNameFilterDisables(not hide)

    def set_filter(self, filter_mode: core.dir.FilterStr | core.QDir.Filter):
        self.setFilter(core.dir.FILTERS.get_enum_value(filter_mode))

    def get_paths(self, indexes: Sequence[core.ModelIndex]) -> list[pathlib.Path]:
        paths = [i.data(self.Roles.FilePathRole) for i in indexes]
        if not paths:
            return []
        if not paths[0]:
            paths = [
                folder / filename
                for folder in paths
                for filename in folder.iterdir()
                if (folder / filename).is_file()
            ]
        return paths

    def get_permissions(
        self, index: core.ModelIndex
    ) -> list[core.filedevice.PermissionStr]:
        return core.filedevice.PERMISSIONS.get_list(self.permissions(index))


class FileSystemModel(
    FileSystemModelMixin, core.AbstractItemModelMixin, widgets.QFileSystemModel
):
    pass


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    app.load_language("de")
    model = FileSystemModel()
    model.set_root_path("root")
    tree = widgets.TreeView()
    tree.set_model(model)
    tree.show()
    app.exec()



# /// script
# requires-python = ">=3.11"
# dependencies = [
#   "prettyqt",
#   "pyside6",
# ]
# ///

import importlib
import pkgutil
import types
from typing import Self

import PyQt6
import PySide6

import prettyqt
from prettyqt import core, gui, widgets
from prettyqt.qt import QtCore, QtGui, QtWidgets


OBSELETE = ["QPictureIO", "QDesktopWidget", "QDirModel"]

mapper = {gui: QtGui, widgets: QtWidgets, core: QtCore}

if __name__ == "__main__":
    for k, v in mapper.items():
        theirs = dir(v)
        ours = dir(k)
        print(f"Missing modules for {k.__name__}")
        for i in theirs:
            if i[1:] not in ours and i[0] == "Q" and i not in OBSELETE:
                print(f"{v.__name__}.{i}")


class API:
    def __init__(self, module: types.ModuleType):
        self.module = module
        self._api = module.__name__

    def qt_name(self, name):
        return name

    def list_modules(self, force_qt_naming=False):
        return [p.name for p in pkgutil.iter_modules(self.module.__path__)]

    def import_modules(self):
        return [importlib.import_module(p) for p in self.list_modules()]

    def compare_with(self, other: Self):
        theirs = dir(other.module)
        ours = dir(self.module)
        print(f"Missing modules for {self.module.__name__}")

        for i in theirs:
            if i not in ours and i not in OBSELETE:
                print(f"{v.__name__}.{i}")


class PrettyQtApi(API):
    MODULE = prettyqt

    def compare_with(self, other: API):
        theirs = dir(other.module)
        ours = dir(self.module)
        print(f"Missing modules for {self.module.__name__}")

        for i in theirs:
            if i[1:] not in ours and i[0] == "Q" and i not in OBSELETE:
                print(f"{v.__name__}.{i}")


class PySide6Api(API):
    MODULE = PyQt6


class PyQt6Api(API):
    MODULE = PySide6



from __future__ import annotations

from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from collections.abc import Iterator

    from prettyqt.qt import QtWebEngineCore


class WebEngineScriptCollection:
    def __init__(self, item: QtWebEngineCore.QWebEngineScriptCollection):
        self.item = item

    def __getattr__(self, val):
        return getattr(self.item, val)

    def __getitem__(self, index: str) -> QtWebEngineCore.QWebEngineScript:
        return self.item.find(index)[0]

    def __len__(self):
        return self.item.count()

    def __iter__(self) -> Iterator[QtWebEngineCore.QWebEngineScript]:
        return iter(self.item.toList())

    def __contains__(self, other: QtWebEngineCore.QWebEngineScript):
        return self.item.contains(other)

    def __add__(
        self,
        other: (
            QtWebEngineCore.QWebEngineScript | list[QtWebEngineCore.QWebEngineScript]
        ),
    ):
        self.item.insert(other)
        return self


if __name__ == "__main__":
    from prettyqt import webenginecore, widgets

    app = widgets.app()
    page = webenginecore.WebEnginePage()
    scripts = page.scripts()
    script = webenginecore.WebEngineScript()
    script.setName("test")
    item = WebEngineScriptCollection(scripts)
    assert not bool(item)
    item += script
    assert script in item
    assert len(item) == 1
    assert bool(item)
    app.exec()



from __future__ import annotations

from typing import TYPE_CHECKING, overload

from prettyqt import core
from prettyqt.utils import listdelegators


if TYPE_CHECKING:
    from collections.abc import Callable

    from prettyqt import widgets


class AnimationGroupMixin(core.AbstractAnimationMixin):
    @overload
    def __getitem__(self, index: int) -> core.QAbstractAnimation: ...

    @overload
    def __getitem__(
        self, index: slice
    ) -> listdelegators.ListDelegator[core.QAbstractAnimation]: ...

    def __getitem__(self, index: int | slice):
        count = self.animationCount()
        match index:
            case int():
                if index < 0:
                    index = count + index
                if index >= count:
                    raise IndexError(index)
                anim = self.animationAt(index)
                if anim is None:
                    raise KeyError(index)
                return anim
            case slice():
                stop = index.stop or count
                rng = range(index.start or 0, stop, index.step or 1)
                anims = [self.animationAt(i) for i in rng]
                return listdelegators.ListDelegator(anims)
            case _:
                raise TypeError(index)

    def __setitem__(self, index: int, value: core.QAbstractAnimation):
        if not (0 <= index < self.animationCount()):
            raise KeyError(index)
        self.takeAnimation(index)
        self.insertAnimation(index, value)

    def __len__(self):
        return self.animationCount()

    def __delitem__(self, index: int):
        if not (0 <= index < self.animationCount()):
            raise KeyError(index)
        self.takeAnimation(index)

    def __add__(self, other: core.QAbstractAnimation):
        self.addAnimation(other)
        return self

    def targetObject(self) -> widgets.QWidget:
        """Return shared targetObject if existing."""
        targets = [
            anim.targetObject()
            for i in range(self.animationCount())
            if isinstance((anim := self.animationAt(i)), core.QPropertyAnimation)
        ]
        if len(targets) != self.animationCount() or len(set(targets)) != 1:
            msg = "Could not find shared targetObject for all animations."
            raise RuntimeError(msg)
        return targets[0]

    def add_property_animation(self, obj: Callable) -> core.PropertyAnimation:
        anim = core.PropertyAnimation()
        anim.apply_to(obj)
        self.addAnimation(anim)
        return anim


class AnimationGroup(AnimationGroupMixin, core.QAnimationGroup):
    """Abstract base class for groups of animations."""


if __name__ == "__main__":
    group = AnimationGroup()
    a1 = core.PauseAnimation()
    a2 = core.PauseAnimation()
    group += a1
    group += a2
    sliced = group[:2]



"""Tests for `prettyqt` package."""

import pytest

from prettyqt import core, webenginecore
from prettyqt.utils import InvalidParamError


def test_webengineurlscheme():
    scheme = webenginecore.WebEngineUrlScheme()
    scheme.set_name("test")
    assert scheme.get_name() == "test"
    scheme.set_syntax("host")
    with pytest.raises(InvalidParamError):
        scheme.set_syntax("test")
    assert scheme.get_syntax() == "host"


def test_webenginehttprequest():
    req = webenginecore.WebEngineHttpRequest()
    headers = {"a": "b"}
    req.set_headers(headers)
    assert req.get_headers() == headers
    req.set_method("post")
    with pytest.raises(InvalidParamError):
        req.set_method("test")
    assert req.get_method() == "post"
    req.set_url("http://www.google.de")
    assert req.get_url() == core.Url("http://www.google.de")
    req.get_post_data()


def test_webengineprofile(qapp):
    profile = webenginecore.WebEngineProfile("MyProfile")
    profile.set_persistent_cookie_policy("force")
    with pytest.raises(InvalidParamError):
        profile.set_persistent_cookie_policy("test")
    assert profile.get_persistent_cookie_policy() == "force"
    profile.set_http_cache_type("disk")
    with pytest.raises(InvalidParamError):
        profile.set_http_cache_type("test")
    assert profile.get_http_cache_type() == "disk"
    profile.get_scripts()


def test_webenginescript():
    script = webenginecore.WebEngineScript()
    script.set_injection_point("document_ready")
    with pytest.raises(InvalidParamError):
        script.set_injection_point("test")
    assert script.get_injection_point() == "document_ready"


def test_webenginescriptcollection():
    page = webenginecore.WebEnginePage()
    script = webenginecore.WebEngineScript()
    script.setName("test")
    item = page.get_scripts()
    assert not bool(item)
    item += script
    assert script in item
    assert len(item) == 1
    assert bool(item)
    for _scr in item:
        pass
    assert item["test"] == script


def test_webenginesettings(qapp):
    page = webenginecore.WebEnginePage()
    settings = page.get_settings()
    settings["auto_load_images"] = False
    assert settings["auto_load_images"] is False
    del settings["auto_load_images"]
    settings.set_unknown_url_scheme_policy("allow_all")
    with pytest.raises(InvalidParamError):
        settings.set_unknown_url_scheme_policy("test")
    assert settings.get_unknown_url_scheme_policy() == "allow_all"
    settings.set_font_family("sans_serif", "verdana")
    with pytest.raises(InvalidParamError):
        settings.set_font_family("test", "test")
    assert settings.get_font_family("sans_serif") == "verdana"
    settings.set_font_size("default_fixed", 14)
    with pytest.raises(InvalidParamError):
        settings.set_font_size("test", "test")
    assert settings.get_font_size("default_fixed") == 14


def test_webenginehistory():
    page = webenginecore.WebEnginePage()
    history = page.get_history()
    assert len(history) < 2  # 0 returns 1 for PySide6 6.5, 0 for PyQt6 6.4
    for _item in history:
        pass
    history.get_items()


def test_webenginepage(qapp):
    page = webenginecore.WebEnginePage()
    page.set_zoom(1.5)
    page.set_url("http://www.google.de")
    assert page.get_url() == core.Url("http://www.google.de")
    page.load_url("http://www.google.de")
    page.find_text("test", backward=True, case_sensitive=True, callback=None)
    page.set_lifecycle_state("discarded")
    with pytest.raises(InvalidParamError):
        page.set_lifecycle_state("test")
    assert page.get_lifecycle_state() == "discarded"
    page.trigger_action("select_all")
    page.set_feature_permission(
        "http://www.google.de", "media_audio_capture", "granted_by_user"
    )
    page.get_settings()
    page.set_setting("auto_load_images", False)
    assert page.get_setting("auto_load_images") is False
    page.get_icon_url()
    page.get_requested_url()
    page.get_scroll_position()
    page.get_contents_size()
    # page.choose_files("open", [], [])



from __future__ import annotations

from prettyqt import gui, widgets


class GraphicsOpacityEffect(widgets.GraphicsEffectMixin, widgets.QGraphicsOpacityEffect):
    def get_opacity_mask(self) -> gui.Brush:
        return gui.Brush(self.opacityMask())



# /// script
# requires-python = ">=3.11"
# ///

import pathlib
import re
import sys


script_path = pathlib.Path(__file__).parent

replacements = []
with (script_path / "enums.txt").open() as f:
    for line in f:
        orig, replacement = line.split()
        orig_re = re.compile(re.escape(orig) + r"(?=\W)")
        replacements.append((orig_re, replacement))


for filename in sys.argv[1:]:
    path = pathlib.Path(filename)
    if path.suffix != ".py":
        continue
    content = path.read_text()
    print(filename)
    for orig_re, replacement in replacements:
        content = orig_re.sub(replacement, content)
    path.write_text(content)



"""Tests for `prettyqt` package."""

from prettyqt import designer


def test_abstractextensionfactory(qtbot):
    designer.AbstractExtensionFactory()


# Segfault with PySide6
# def test_pydesignertaskmenuextension(qtbot):
#     parent = core.Object()
#     designer.PyDesignerTaskMenuExtension(parent)



from __future__ import annotations

import logging
import re
import webbrowser

from prettyqt import core, widgets
from prettyqt.custom_network import downloadworker


logger = logging.getLogger(__name__)

GITHUB_MESSAGE = (
    "You can download the new version for your operating system from GitHub.\n\n"
    "Do you want to visit the release website now?"
)

PYPI_MESSAGE = (
    "You should upgrade from command line with 'pip install prettyqt --upgrade'.\n\n"
    "Do you want to view the changelog on github?"
)


class UpdateChecker(core.Object):
    """Check for a new prettyqt version."""

    version_checked = core.Signal(str)  # newest available version
    update_button_clicked = core.Signal(str)  # url to new version
    download_finished = core.Signal(bytes, str)  # response, url
    download_failed = core.Signal(str, str)  # m

    def __init__(
        self,
        app_name: str | None = None,
        app_version: str | None = None,
        releases_url: str | None = None,
        pypi_url: str | None = None,
        changelog_url: str | None = None,
        website_url: str | None = None,
        packaged: bool = False,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.packaged = packaged
        self.releases_url = releases_url
        self.pypi_url = pypi_url
        self.changelog_url = changelog_url
        self.website_url = website_url
        self.current_v = app_version or widgets.app().applicationVersion()
        self.app_name = app_name or widgets.app().applicationName()
        self.threadpool = core.ThreadPool()
        self.download_finished.connect(self._on_download_finished)
        self.message_box = widgets.MessageBox()
        self.message_box.set_icon("mdi.download")
        self.message_box.set_standard_buttons(["ok", "cancel"])

        self.message_box.set_default_button("ok")
        if self.changelog_url:
            self.message_box.add_custom_button(
                "Show changelog",
                "accept",
                callback=lambda: webbrowser.open(self.changelog_url),
            )

    def get(self, url: str, timeout: int = 30) -> None:
        logger.debug("Downloading %s...", url)
        worker = downloadworker.DownloadWorker(url=url, timeout=timeout)
        worker.signals.download_finished.connect(self.download_finished)
        worker.signals.download_failed.connect(self.download_failed)
        self.threadpool.start(worker)

    @core.Slot(bytes, str)
    def _on_download_finished(self, data: bytes, url: str) -> None:
        """Parse the tag version from the response and emit version retrieved signal."""
        newest_v = None
        try:
            text = data.decode(errors="ignore")
            if self.packaged:
                regex = r"/releases/tag/v(\d+\.\d+\.\d+)\""  # atom
            else:
                regex = r"\"version\":\s*\"(\d+\.\d+\.\d+)\""  # json

            match = re.search(regex, text)
            if match and match[1]:
                newest_v = match[1]
        except Exception as e:
            logger.exception("Parsing response of update check failed: %s", e)

        if not newest_v:
            logger.error("Could not detect remote version. Update check won't work!")
            return
        logger.debug("Newest version: %s (installed: %s)", newest_v, self.current_v)
        self.version_checked.emit(newest_v)
        if core.VersionNumber(newest_v) > core.VersionNumber(self.current_v):
            self._show_update_dialog(new_version=newest_v)

    def _show_update_dialog(self, new_version: str) -> None:
        """Show dialog informing about available update."""
        text = (
            f"<b>{self.app_name} {new_version} is available.</b> "
            f"(Current version: {self.current_v})"
        )
        self.message_box.setText(text)

        info_text = GITHUB_MESSAGE if self.packaged else PYPI_MESSAGE
        self.message_box.setInformativeText(info_text)
        self.message_box.set_cursor("arrow")

        if self.message_box.show_blocking() == "ok":
            update_url = self.releases_url if self.packaged else self.changelog_url
            self.update_button_clicked.emit(update_url)

    def check_for_update(self) -> None:
        """Start the update check."""
        url = f"{self.releases_url}.atom" if self.packaged else f"{self.pypi_url}/json"
        logger.debug("Search for new version on %s", url)
        self.get(url, timeout=10)


if __name__ == "__main__":
    app = widgets.app()
    checker = UpdateChecker(
        app_version="1.0.0",
        releases_url="https://github.com/phil65/prettyqt/releases",
        changelog_url="https://github.com/phil65/prettyqt/blob/main/CHANGELOG",
        pypi_url="https://pypi.org/pypi/prettyqt",
        website_url="https://github.com/phil65/prettyqt",
        parent=app,
    )
    checker.check_for_update()
    app.exec()



from __future__ import annotations

from prettyqt import core
from prettyqt.qt import QtHelp


class HelpIndexModel(core.StringListModelMixin, QtHelp.QHelpIndexModel):
    """Model that supplies index keywords to views."""



from __future__ import annotations

from typing import TYPE_CHECKING, Self

from prettyqt import core


if TYPE_CHECKING:
    from collections.abc import Callable, Iterator

    from prettyqt.utils import datatypes


class VariantAnimationMixin(core.AbstractAnimationMixin):
    def __getitem__(self, value: float) -> datatypes.Variant:
        return self.keyValueAt(value)

    def __setitem__(self, key: float, value: datatypes.Variant):
        self.setKeyValueAt(key, value)

    def __iter__(self) -> Iterator[tuple[float, datatypes.Variant]]:
        return iter(self.keyValues())

    def set_easing(
        self,
        easing_type: (
            core.easingcurve.TypeStr | core.QEasingCurve.Type | Callable[[float], float]
        ),
    ):
        curve = core.EasingCurve()
        if callable(easing_type):
            curve.set_custom_type(easing_type)
        else:
            curve.set_type(easing_type)
        self.setEasingCurve(curve)

    def get_easing(self) -> core.easingcurve.TypeStr | Callable[[float], float]:
        curve = core.EasingCurve(self.easingCurve())
        typ = curve.get_type()
        return curve.get_custom_type() if typ == "custom" else typ

    def set_range(self, start, end):
        self.setStartValue(start)
        self.setEndValue(end)

    def reverse(self):
        """True reverse instead of just setting direction."""
        self.setKeyValues(list(reversed(self.keyValues())))

    def reversed(self) -> VariantAnimation:
        """Return a reversed copy of the animation."""
        new = self.get_metaobject().copy(self)
        new.reverse()
        return new

    def append_reversed(self) -> Self:
        """Append the reversed animation, effectively doubling the duration."""
        self.setDuration(self.duration() * 2)
        first_part = [(k / 2, v) for k, v in self.keyValues()]
        second_part = [(1 - (k / 2), v) for k, v in self.keyValues()]
        keys = first_part + list(reversed(second_part))[1:]
        self.setKeyValues(keys)
        return self


class VariantAnimation(VariantAnimationMixin, core.QVariantAnimation):
    """Base class for animations."""


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    anim = VariantAnimation()
    anim.set_start_value(10)
    anim = anim.reversed()
    anim.run()
    print(anim.endValue())



from __future__ import annotations

from prettyqt import core
from prettyqt.qt import QtGui
from prettyqt.utils import get_repr


class ValidatorMixin(core.ObjectMixin):
    def __repr__(self):
        return get_repr(self)

    def __and__(self, other: Validator):
        from prettyqt import validators

        return validators.AndValidator([self, other])

    def __or__(self, other: Validator):
        from prettyqt import validators

        return validators.OrValidator([self, other])

    def is_valid_value(self, value: str, pos: int = 0) -> bool:
        val = self.validate(value, pos)
        return val[0] == self.State.Acceptable  # type: ignore


class Validator(ValidatorMixin, QtGui.QValidator):
    pass


if __name__ == "__main__":
    val = Validator() & Validator()
    print(val)



from __future__ import annotations

from typing import Literal

from prettyqt import gui
from prettyqt.utils import bidict


PDF_VERSION = bidict(
    v1_4=gui.QPagedPaintDevice.PdfVersion.PdfVersion_1_4,
    va1b=gui.QPagedPaintDevice.PdfVersion.PdfVersion_A1b,
    v1_6=gui.QPagedPaintDevice.PdfVersion.PdfVersion_1_6,
)

PdfVersionStr = Literal["v1_4", "va1b", "v1_6"]


class PagedPaintDeviceMixin(gui.PaintDeviceMixin):
    def get_page_ranges(self) -> list[gui.PageRanges]:
        return [gui.PageRanges(i) for i in self.pageRanges()]

    def get_page_layout(self) -> gui.PageLayout:
        return gui.PageLayout(self.PageLayout())

    def set_page_ranges(self, ranges: gui.QPageRanges | list[tuple[int, int]]):
        if isinstance(ranges, gui.QPageRanges):
            self.setPageRanges(ranges)
        else:
            ranges = gui.PageRanges()
            for start, end in ranges:
                ranges.addRange(start, end)

    def set_page_orientation(self, orientation: gui.pagelayout.OrientationStr):
        self.setPageOrientation(gui.pagelayout.ORIENTATIONS[orientation])


class PagedPaintDevice(PagedPaintDeviceMixin, gui.QPagedPaintDevice):
    pass


if __name__ == "__main__":
    device = PagedPaintDevice()



"""Tests for `prettyqt` package."""

import pytest


bluetooth = pytest.importorskip("prettyqt.bluetooth")

# from prettyqt.utils import InvalidParamError


def test_bluetoothaddress():
    address = bluetooth.BluetoothAddress()
    repr(address)
    assert not address


def test_bluetoothservicediscoveryagent(qtlog):
    with qtlog.disabled():
        agent = bluetooth.BluetoothServiceDiscoveryAgent()
    assert agent.get_error() == "none"
    assert agent.get_discovered_services() == []


def test_bluetoothdevicediscoveryagent(qtlog):
    with qtlog.disabled():
        agent = bluetooth.BluetoothDeviceDiscoveryAgent()
    assert agent.get_error() == "none"
    agent.get_supported_discovery_methods()



from __future__ import annotations

import os

from prettyqt import core, gui, widgets


class CrossFadeWidget(widgets.Widget):
    def __init__(
        self,
        pixmap_1: gui.QPixmap | os.PathLike[str] | str | None = None,
        pixmap_2: gui.QPixmap | os.PathLike[str] | str | None = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.pixmap_1 = None
        self.pixmap_2 = None
        self.set_pixmap_1(pixmap_1)
        self.set_pixmap_2(pixmap_2)
        self.set_blend_factor(0.0)
        self.timeline = core.TimeLine(finished=self.close)
        self.timeline.value_changed.connect(self.set_blend_factor)
        self.set_size_policy("fixed", "fixed")

    def start_fade(self, duration: int = 500):
        self.timeline.setDuration(duration)
        self.timeline.start()
        # self.show()

    def set_pixmap_1(
        self, pixmap: gui.QPixmap | os.PathLike[str] | str | widgets.QWidget | None
    ):
        self.pixmap_1 = self._get_pixmap(pixmap)
        self.updateGeometry()
        self.repaint()

    def set_pixmap_2(
        self, pixmap: gui.QPixmap | os.PathLike[str] | str | widgets.QWidget | None
    ):
        self.pixmap_2 = self._get_pixmap(pixmap)
        self.updateGeometry()
        self.repaint()

    @staticmethod
    def _get_pixmap(
        pixmap: gui.QPixmap | os.PathLike[str] | str | widgets.QWidget | None,
    ):
        match pixmap:
            case gui.QPixmap():
                return pixmap
            case os.PathLike():
                return gui.Pixmap(pixmap)
            case widgets.QWidget():
                return pixmap.grab()
            case None:
                return gui.Pixmap()

    def set_blend_factor(self, factor: float):
        self._blend_factor = factor
        self.repaint()

    def get_blend_factor(self) -> float:
        """Pixmap blending factor between 0.0 and 1.0."""
        return self._blend_factor

    def sizeHint(self):
        """Return max size of both."""
        size1 = self.pixmap_1.size() if self.pixmap_1 else super().sizeHint()
        size2 = self.pixmap_2.size() if self.pixmap_2 else super().sizeHint()
        return size1.expandedTo(size2)

    def paintEvent(self, event):
        """Paint the interpolated pixmap image."""
        with gui.Painter(self) as p:
            p.setClipRect(event.rect())
            factor = self.get_blend_factor() ** 2
            if self.pixmap_1 and 1.0 - factor:
                p.setOpacity(1.0 - factor)
                p.drawPixmap(core.Point(0, 0), self.pixmap_1)
            if self.pixmap_2 and factor:
                p.setOpacity(factor)
                p.drawPixmap(core.Point(0, 0), self.pixmap_2)

    blend_factor = core.Property(
        float,
        get_blend_factor,
        set_blend_factor,
        doc="Blend factor between the two pixmaps.",
    )


if __name__ == "__main__":
    app = widgets.app()
    w1 = widgets.RadioButton("fdsfsdf")
    w2 = widgets.PlainTextEdit("fbbb")
    widget = CrossFadeWidget(w1, w2)
    container = widgets.Widget()
    container.set_layout("vertical")
    container.box.add(widget)
    slider = widgets.Slider("horizontal", maximum=100)
    container.box.add(slider)
    slider.valueChanged.connect(lambda x: widget.set_blend_factor(x / 100))
    container.show()
    with app.debug_mode():
        app.exec()



"""Tests for `prettyqt` package."""

from prettyqt import itemdelegates, widgets
from prettyqt.qt import QtCore


def test_buttondelegate(qtbot):
    table = widgets.TableView()
    widget = itemdelegates.ButtonDelegate(parent=table)
    widget.setEditorData(widgets.Widget(), None)
    widget.createEditor(None, None, QtCore.QModelIndex())
    # widget.currentIndexChanged()


def test_radiodelegate(qtbot, tablewidget):
    delegate = itemdelegates.RadioDelegate(tablewidget, ["a", "b"])
    tablewidget.show()
    tablewidget.setItemDelegateForColumn(0, delegate)
    tablewidget.openPersistentEditor(tablewidget[0, 0])
    tablewidget.hide()


def test_icondelegate(qtbot, tablewidget):
    delegate = itemdelegates.IconDelegate()
    tablewidget.show()
    tablewidget.setItemDelegateForColumn(0, delegate)
    tablewidget.hide()


def test_nofocusdelegate(qtbot, tablewidget):
    delegate = itemdelegates.NoFocusDelegate()
    tablewidget.show()
    tablewidget.setItemDelegateForColumn(0, delegate)
    tablewidget.hide()


def test_stardelegate(qtbot, tablewidget):
    delegate = itemdelegates.StarDelegate()
    tablewidget.show()
    tablewidget.setItemDelegateForColumn(0, delegate)
    tablewidget.hide()


def test_progressbardelegate(qtbot, tablewidget):
    delegate = itemdelegates.ProgressBarDelegate()
    tablewidget.show()
    tablewidget.setItemDelegateForColumn(0, delegate)
    tablewidget.hide()


def test_editordelegate(qtbot, tablewidget):
    delegate = itemdelegates.EditorDelegate(parent=tablewidget)
    tablewidget.show()
    tablewidget.setItemDelegateForColumn(0, delegate)
    tablewidget.hide()


def test_widgetdelegate(qtbot, tablewidget):
    delegate = itemdelegates.EditorDelegate(parent=tablewidget)
    tablewidget.show()
    tablewidget.setItemDelegateForColumn(0, delegate)
    tablewidget.hide()



from __future__ import annotations

import io
import mimetypes
import os
from typing import Literal, Self

from prettyqt import core, gui
from prettyqt.qt import API
from prettyqt.utils import bidict, helpers, serializemixin


FormatStr = Literal[
    "invalid",
    "mono",
    "monolsb",
    "indexed8",
    "rgb32",
    "argb32",
    "argb32_premultiplied",
    "rgb16",
    "argb8565_premultiplied",
    "rgb666",
    "argb6666_premultiplied",
    "rgb555",
    "argb8555_premultiplied",
    "rgb888",
    "rgb444",
    "argb4444_premultiplied",
    "rgbx8888",
    "rgba8888",
    "rgba8888_premultiplied",
    "bgr30",
    "a2bgr30_premultiplied",
    "rgb30",
    "a2rgb30_premultiplied",
    "alpha8",
    "grayscale8",
    "grayscale16",
    "rgbx64",
    "rgba64",
    "rgba64_premultiplied",
    "bgr888",
    "rgbx16fpx4",
    "rgba16fpx4",
    "rgba16fpx4_premultiplied",
    "rgbx32fpx4",
    "rgba32fpx4",
    "rgba32fpx4_premultiplied",
]

FORMAT: bidict[FormatStr, gui.QImage.Format] = bidict(
    invalid=gui.QImage.Format.Format_Invalid,
    mono=gui.QImage.Format.Format_Mono,
    monolsb=gui.QImage.Format.Format_MonoLSB,
    indexed8=gui.QImage.Format.Format_Indexed8,
    rgb32=gui.QImage.Format.Format_RGB32,
    argb32=gui.QImage.Format.Format_ARGB32,
    argb32_premultiplied=gui.QImage.Format.Format_ARGB32_Premultiplied,
    rgb16=gui.QImage.Format.Format_RGB16,
    argb8565_premultiplied=gui.QImage.Format.Format_ARGB8565_Premultiplied,
    rgb666=gui.QImage.Format.Format_RGB666,
    argb6666_premultiplied=gui.QImage.Format.Format_ARGB6666_Premultiplied,
    rgb555=gui.QImage.Format.Format_RGB555,
    argb8555_premultiplied=gui.QImage.Format.Format_ARGB8555_Premultiplied,
    rgb888=gui.QImage.Format.Format_RGB888,
    rgb444=gui.QImage.Format.Format_RGB444,
    argb4444_premultiplied=gui.QImage.Format.Format_ARGB4444_Premultiplied,
    rgbx8888=gui.QImage.Format.Format_RGBX8888,
    rgba8888=gui.QImage.Format.Format_RGBA8888,
    rgba8888_premultiplied=gui.QImage.Format.Format_RGBA8888_Premultiplied,
    bgr30=gui.QImage.Format.Format_BGR30,
    a2bgr30_premultiplied=gui.QImage.Format.Format_A2BGR30_Premultiplied,
    rgb30=gui.QImage.Format.Format_RGB30,
    a2rgb30_premultiplied=gui.QImage.Format.Format_A2RGB30_Premultiplied,
    alpha8=gui.QImage.Format.Format_Alpha8,
    grayscale8=gui.QImage.Format.Format_Grayscale8,
    grayscale16=gui.QImage.Format.Format_Grayscale16,
    rgbx64=gui.QImage.Format.Format_RGBX64,
    rgba64=gui.QImage.Format.Format_RGBA64,
    rgba64_premultiplied=gui.QImage.Format.Format_RGBA64_Premultiplied,
    bgr888=gui.QImage.Format.Format_BGR888,
    rgbx16fpx4=gui.QImage.Format.Format_RGBX16FPx4,
    rgba16fpx4=gui.QImage.Format.Format_RGBA16FPx4,
    rgba16fpx4_premultiplied=gui.QImage.Format.Format_RGBA16FPx4_Premultiplied,
    rgbx32fpx4=gui.QImage.Format.Format_RGBX32FPx4,
    rgba32fpx4=gui.QImage.Format.Format_RGBA32FPx4,
    rgba32fpx4_premultiplied=gui.QImage.Format.Format_RGBA32FPx4_Premultiplied,
)


class Image(serializemixin.SerializeMixin, gui.PaintDeviceMixin, gui.QImage):
    """Hardware-independent image representation with direct access to the pixel data."""

    def __setitem__(self, index: tuple[int, int], value):
        self.setPixel(index[0], index[1], value)

    def __getitem__(self, index: tuple[int, int]) -> int | list[int]:
        match index:
            case int() as row, int() as col:
                return self.pixel(row, col)
            case (row, col):
                rowcount = self.height()
                colcount = self.width()
                return [
                    self.pixel(i, j)
                    for i, j in helpers.iter_positions(row, col, rowcount, colcount)
                ]
            case _:
                raise TypeError(index)

    @classmethod
    def from_ndarray(cls, arr) -> Self:
        import numpy as np

        height, width, bytes_per_component = arr.shape
        if arr.dtype in {np.float32, np.float64}:
            arr = (255 * arr).round()
        arr = arr.astype(np.uint8)
        return cls(
            arr.data,
            width,
            height,
            bytes_per_component * width,
            gui.QImage.Format.Format_RGB888,
        )

    def to_ndarray(
        self, fmt: FormatStr | gui.QImage.Format = "rgb888", channels: int = 3
    ):
        import numpy as np

        qimage = self.convert_to_format(fmt)
        width = qimage.width()
        height = qimage.height()

        ptr = qimage.constBits()
        return np.array(ptr).reshape(height, width, channels)  # Copies the data

    @classmethod
    def from_pil(cls, image) -> Self:
        # from https://github.com/python-pillow/Pillow/blob/main/src/PIL/ImageQt.py
        from PIL import ImageQt

        data = ImageQt._toqclass_helper(image)
        img = cls(data["data"], data["size"][0], data["size"][1], data["format"])
        if data["colortable"]:
            img.setColorTable(data["colortable"])
        img.__data = data["data"]
        return img

    @classmethod
    def for_mimetype(cls, path: os.PathLike[str] | str) -> Self | None:
        """Try to create an icon from theme using the file mimetype.

        E.g.::

            return self.mimetype_icon(path, fallback=":/icons/text-x-python.png")

        :param path: file path for which the icon must be created
        :param fallback: fallback icon path (qrc or file system)
        :returns: QIcon or None if the file mimetype icon could not be found.
        """
        path = os.fspath(path)
        if mime := mimetypes.guess_type(path)[0]:
            icon = mime.replace("/", "-")
            # if system.WINDOWS:
            #     return icons.file()
            if cls.hasThemeIcon(icon):
                icon = cls(cls.fromTheme(icon))
                if not icon.isNull():
                    return icon
        return None  # gui.Icon.fromTheme("text-x-generic")

    def to_pil(self):
        from PIL import Image as PILImage

        buffer = core.Buffer()
        buffer.open(core.Buffer.OpenModeFlag.ReadWrite)
        self.save(buffer, "PNG")
        return PILImage.open(io.BytesIO(buffer.data()))

    def invert_pixels(self, invert_alpha: bool = False):
        self.invertPixels(
            gui.QImage.InvertMode.InvertRgba
            if invert_alpha
            else gui.QImage.InvertMode.InvertRgb
        )

    def convert_to_format(self, fmt: FormatStr | gui.QImage.Format) -> Self:
        return type(self)(self.convertToFormat(FORMAT.get_enum_value(fmt)))

    def as_bytes(self) -> bytes:
        bits = self.bits()
        if bits is None:
            return b""
        match API:
            case "pyqt6":
                return bits.asstring(self.sizeInBytes())
            case "pyside6":
                return bits.tobytes()


if __name__ == "__main__":
    app = gui.app()
    image = gui.Pixmap(100, 100).toImage()
    image = Image(image)
    print(len(bytes(image)))



from __future__ import annotations

from typing import Literal

from prettyqt import statemachine
from prettyqt.utils import bidict


HistoryTypeStr = Literal["shallow", "deep"]

HISTORY_TYPE: bidict[HistoryTypeStr, statemachine.QHistoryState.HistoryType] = bidict(
    shallow=statemachine.QHistoryState.HistoryType.ShallowHistory,
    deep=statemachine.QHistoryState.HistoryType.DeepHistory,
)


class HistoryState(statemachine.AbstractStateMixin, statemachine.QHistoryState):
    def set_history_type(
        self, typ: HistoryTypeStr | statemachine.QHistoryState.HistoryType
    ):
        """Set history type to use.

        Args:
            typ: history type to use
        """
        self.setHistoryType(HISTORY_TYPE.get_enum_value(typ))

    def get_history_type(self) -> HistoryTypeStr:
        """Return current history type.

        Returns:
            history type
        """
        return HISTORY_TYPE.inverse[self.historyType()]


if __name__ == "__main__":
    state = HistoryState()



"""Classes to make printing easier and more portable."""

from __future__ import annotations

from prettyqt.qt.QtPrintSupport import *  # noqa: F403

from .printer import Printer
from .abstractprintdialog import AbstractPrintDialog, AbstractPrintDialogMixin
from .printdialog import PrintDialog
from .pagesetupdialog import PageSetupDialog
from .printengine import PrintEngine
from .printpreviewwidget import PrintPreviewWidget
from .printpreviewdialog import PrintPreviewDialog
from prettyqt.qt import QtPrintSupport

QT_MODULE = QtPrintSupport

__all__ = [
    "AbstractPrintDialog",
    "AbstractPrintDialogMixin",
    "PageSetupDialog",
    "PrintDialog",
    "PrintEngine",
    "PrintPreviewDialog",
    "PrintPreviewWidget",
    "Printer",
]



from __future__ import annotations

from collections.abc import Callable
import functools
import inspect
import logging
from typing import Self

from prettyqt import core


logger = logging.getLogger(__name__)


class Signals(core.Object):
    finished = core.Signal()
    error = core.Signal(Exception)
    result = core.Signal(object)
    progress = core.Signal(int)


class Worker(core.Runnable):
    """Worker thread.

    Inherits from QRunnable to handler worker thread setup, signals and wrap-up.

    Arguments:
        fn:     The function callback to run on this worker thread.
                Supplied args and kwargs will be passed through to the runner.
        args:   Arguments to pass to the callback function
        kwargs: Keywords to pass to the callback function

    """

    def __init__(self, fn: Callable, *args, **kwargs):
        self.fn = fn
        self.args = args
        self.keyword_args = kwargs
        self.signals = Signals()
        super().__init__()

        # Add the callback to our kwargs
        # self.keyword_args["progress_callback"] = self.signals.progress

    def __getattr__(self, name: str) -> core.SignalInstance:
        attr = getattr(self.signals, name, None)
        if isinstance(attr, core.SignalInstance):
            return getattr(self.signals, name)
        msg = f"{type(self).__name__!r} has no attribute {name!r}"
        raise AttributeError(msg)

    @core.Slot()
    def run(self):
        """Initialise the runner function with passed args, kwargs."""
        try:
            if inspect.isgeneratorfunction(self.fn):
                for result in self.fn(*self.args, **self.keyword_args):
                    self.signals.result.emit(result)
            else:
                result = self.fn(*self.args, **self.keyword_args)
                self.signals.result.emit(result)
            self.signals.finished.emit()
        except Exception as e:
            logger.exception("Error occurred in function %r", self.fn.__name__)
            self.signals.error.emit(e)


def run_async(
    result_fn: Callable | None = None,
    finished_fn: Callable | None = None,
    progress_fn: Callable | None = None,
    error_fn: Callable | None = None,
):
    def inner(func: Callable) -> Callable:
        @functools.wraps(func)
        def async_func(*args, **kwargs):
            pool = ThreadPool.instance()
            pool.start_worker(
                func,
                args,
                kwargs,
                result_fn=result_fn,
                finished_fn=finished_fn,
                progress_fn=progress_fn,
                error_fn=error_fn,
            )

        return async_func

    return inner


class ThreadPool(core.ObjectMixin, core.QThreadPool):
    """Manages a collection of QThreads.

    Note: signals only work correctly when exclusively using start_worker method.
    """

    __instance: Self | None = None  # a global instance

    job_num_updated = core.Signal(int)
    error_occured = core.Signal(Exception)
    busy_state_changed = core.Signal(bool)

    def __contains__(self, other: core.QThread):
        return self.contains(other)

    def get_thread_priority(self) -> core.thread.PriorityStr:
        return core.thread.PRIORITY.inverse[self.threadPriority()]

    def set_thread_priority(self, priority: core.thread.PriorityStr):
        prio = core.thread.PRIORITY[priority]
        self.setThreadPriority(prio)

    @classmethod
    def instance(cls) -> Self:
        """Return global ThreadPool singleton. (globalInstance always returns Qt type)."""
        if cls.__instance is None:
            cls.__instance = cls()
        return cls.__instance

    def start_worker(
        self,
        fn_or_worker: Callable | Worker,
        args: tuple | None = None,
        kwargs: dict | None = None,
        priority: int = 0,
        result_fn: Callable | None = None,
        finished_fn: Callable | None = None,
        progress_fn: Callable | None = None,
        error_fn: Callable | None = None,
    ):
        if isinstance(fn_or_worker, Callable):
            if args is None:
                args = ()
            if kwargs is None:
                kwargs = {}
            runnable = Worker(fn_or_worker, *args, **kwargs)
        else:
            runnable = fn_or_worker
        runnable.signals.finished.connect(self._on_job_ended)
        runnable.signals.error.connect(self._on_job_ended)
        runnable.signals.error.connect(self._on_exception)
        if result_fn:
            runnable.signals.result.connect(result_fn)
        if finished_fn:
            runnable.signals.finished.connect(finished_fn)
        if progress_fn:
            runnable.signals.progress.connect(progress_fn)
        if error_fn:
            runnable.signals.error.connect(error_fn)
        thread_count = self.activeThreadCount()
        if thread_count == 0:
            self.busy_state_changed.emit(True)
        self.job_num_updated.emit(thread_count + 1)  # + 1 because we didnt start yet.
        super().start(runnable, priority)

    def _on_job_ended(self):
        thread_count = self.activeThreadCount()
        if thread_count == 1:  # this is the last job
            self.busy_state_changed.emit(False)
        self.job_num_updated.emit(thread_count - 1)  # -1 because we didnt really end yet.

    def _on_exception(self, exception):
        self.error_occured.emit(exception)


if __name__ == "__main__":
    pool = ThreadPool()



from __future__ import annotations

from typing import Literal

from prettyqt import core, gui
from prettyqt.utils import bidict


ModeStr = Literal["clipboard", "selection", "find_buffer"]

MODES: bidict[ModeStr, gui.QClipboard.Mode] = bidict(
    clipboard=gui.QClipboard.Mode.Clipboard,
    selection=gui.QClipboard.Mode.Selection,
    find_buffer=gui.QClipboard.Mode.FindBuffer,
)


class Clipboard(core.ObjectMixin):
    """Access to the window system clipboard."""

    def __init__(self, item: gui.QClipboard):
        self.item = item

    def __getattr__(self, val):
        return getattr(self.item, val)

    def set_pixmap(
        self,
        pixmap: gui.QPixmap | None,
        mode: ModeStr | gui.QClipboard.Mode = "clipboard",
    ):
        if pixmap is None:
            pixmap = gui.QPixmap()
        self.item.setPixmap(pixmap, MODES.get_enum_value(mode))

    def get_pixmap(
        self, mode: ModeStr | gui.QClipboard.Mode = "clipboard"
    ) -> gui.Pixmap | None:
        val = MODES.get_enum_value(mode)
        pix = gui.Pixmap(self.item.pixmap(val))
        return None if pix.isNull() else pix

    def set_image(
        self,
        image: gui.QImage | None,
        mode: ModeStr | gui.QClipboard.Mode = "clipboard",
    ):
        if image is None:
            image = gui.QImage()
        self.item.setImage(image, MODES.get_enum_value(mode))

    def get_image(
        self, mode: ModeStr | gui.QClipboard.Mode = "clipboard"
    ) -> gui.Image | None:
        img = gui.Image(self.item.image(MODES.get_enum_value(mode)))
        return None if img.isNull() else img

    def set_mimedata(
        self,
        mimedata: core.QMimeData,
        mode: ModeStr | gui.QClipboard.Mode = "clipboard",
    ):
        self.item.setMimeData(mimedata, MODES.get_enum_value(mode))

    def get_mimedata(
        self, mode: ModeStr | gui.QClipboard.Mode = "clipboard"
    ) -> core.QMimeData:
        return self.item.mimeData(MODES.get_enum_value(mode))

    def set_text(self, text: str, mode: ModeStr | gui.QClipboard.Mode = "clipboard"):
        self.item.setText(text, MODES.get_enum_value(mode))

    def get_text(self, mode: ModeStr | gui.QClipboard.Mode = "clipboard") -> str:
        return self.item.text(MODES.get_enum_value(mode))



"""Tests for `prettyqt` package."""

import inspect
import pathlib
import pickle

from prettyqt import validators


clsmembers = inspect.getmembers(validators, inspect.isclass)


def pickle_roundtrip(something):
    path = pathlib.Path("data.pkl")
    with path.open("wb") as jar:
        pickle.dump(something, jar)
    with path.open("rb") as jar:
        return pickle.load(jar)


# @pytest.mark.parametrize("name, cls", clsmembers)
# def test_pickle(name, cls):
#     vale = cls()
#     with open("data.pkl", "wb") as jar:
#         pickle.dump(vale, jar)
#     with open("data.pkl", "rb") as jar:
#         vale = pickle.load(jar)


def test_pathvalidator():
    val = validators.PathValidator()
    val = pickle_roundtrip(val)
    assert val.is_valid_value(str(pathlib.Path.cwd()))
    assert not val.is_valid_value("abs")
    repr(val)
    assert val == validators.PathValidator()
    assert val != validators.NotZeroValidator()


def test_notzerovalidator():
    val = validators.NotZeroValidator()
    assert val.is_valid_value("1")
    assert not val.is_valid_value("0")
    repr(val)
    assert val != validators.PathValidator()
    assert val == validators.NotZeroValidator()


def test_notemptyvalidator():
    val = validators.NotEmptyValidator()
    assert val.is_valid_value("1")
    assert not val.is_valid_value("")
    repr(val)
    assert val != validators.PathValidator()
    assert val == validators.NotEmptyValidator()


def test_andvalidator():
    val1 = validators.NotEmptyValidator()
    val2 = validators.NotZeroValidator()
    composite = validators.AndValidator([val1, val2])
    assert composite != validators.PathValidator()
    assert composite == validators.AndValidator([val1, val2])
    composite = pickle_roundtrip(composite)
    assert composite.is_valid_value("1")
    assert not composite.is_valid_value("")
    assert not composite.is_valid_value("0")
    val1 & val2
    for _child in composite:
        pass
    assert val1 in composite
    assert val1 == composite[0]
    composite[1] = validators.NotZeroValidator()
    assert len(composite) == 2
    del composite[1]
    repr(composite)


def test_intlistvalidator():
    val = validators.IntListValidator(allow_single=True)
    assert val.is_valid_value("1")
    assert val.is_valid_value("1,2")
    val = validators.IntListValidator(allow_single=False)
    assert not val.is_valid_value("1")
    assert val.is_valid_value("1,2")
    repr(val)
    assert val != validators.PathValidator()
    assert val == validators.IntListValidator(allow_single=False)


def test_floatlistvalidator():
    val = validators.FloatListValidator(allow_single=True)
    assert val.is_valid_value("1.0")
    assert val.is_valid_value("1.0,2")
    val = validators.FloatListValidator(allow_single=False)
    assert not val.is_valid_value("1.0")
    assert val.is_valid_value("1.0,2")
    repr(val)
    assert val != validators.PathValidator()
    assert val == validators.FloatListValidator(allow_single=False)


def test_regexpatternvalidator():
    val = validators.RegexPatternValidator()
    assert val.is_valid_value("[") is False
    assert val.is_valid_value("[0-9]") is True
    repr(val)
    assert val != validators.PathValidator()
    assert val == validators.RegexPatternValidator()



from __future__ import annotations

from typing import Literal

from prettyqt import constants, core, widgets
from prettyqt.qt import QtGui
from prettyqt.utils import datatypes, get_repr, listdelegators


# from typing import Any


class ToolBarMixin(widgets.WidgetMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.set_icon_size(24)
        self._menu_buttons = []
        self._tooltip_labels = []
        self._tooltip_timer = core.Timer(timeout=self.hide_tooltips)

    def __getitem__(
        self, row: int | slice
    ) -> widgets.QWidget | listdelegators.ListDelegator[widgets.QWidget]:
        match row:
            case int():
                action = self.actions()[row]
                if action is None:
                    raise KeyError(row)
                return self.widgetForAction(action)
            case slice():
                actions = self.actions()
                count = len(actions) if row.stop is None else row.stop
                values = list(range(count)[row])
                actions = [actions[i] for i in values]
                ls = [self.widgetForAction(i) for i in actions]
                return listdelegators.ListDelegator(ls)
            case _:
                raise TypeError(row)

    # def __setstate__(self, state: dict[str, Any]) -> None:
    #     super().__setstate__(state)
    #     self.addActions(state["actions"])

    # def __reduce__(self):
    #     return type(self), (), self.__getstate__()

    # def serialize_fields(self):
    #     return dict(actions=self.actions(), allowed_areas=self.get_allowed_areas())

    def __repr__(self):
        return get_repr(self, self.windowTitle())

    def _get_map(self):
        maps = super()._get_map()
        maps |= {
            "toolButtonStyle": constants.TOOLBUTTON_STYLE,
            "allowedAreas": constants.TOOLBAR_AREA,
        }
        return maps

    def add(self, item: QtGui.QAction | widgets.QWidget):
        if isinstance(item, QtGui.QAction):
            self.addAction(item)
        else:
            self.addWidget(item)

    def add_menu_button(
        self, label: str, icon: datatypes.IconType, menu: widgets.QMenu
    ) -> widgets.ToolButton:
        btn = widgets.ToolButton.for_menu(menu, icon=icon)
        btn.setText(label)
        btn.setToolButtonStyle(self.toolButtonStyle())
        self._menu_buttons.append(btn)
        self.addWidget(btn)
        return btn

    def add_separator(
        self, text: str | None = None, before: QtGui.QAction | None = None
    ) -> QtGui.QAction:
        """Adds a separator showing an optional label.

        Args:
            text: Text to show on separator
            before: insert separator before specific action

        Returns:
            Separator action
        """
        if text is None:
            return self.insertSeparator(before) if before else self.addSeparator()
        label = widgets.Label(text)
        label.setMinimumWidth(self.minimumWidth())
        with label.edit_stylesheet() as ss:
            ss.background.setValue("lightgrey")
        label.set_alignment(horizontal="center")
        return self.insertWidget(before, label) if before else self.addWidget(label)

    def set_style(self, style: constants.ToolButtonStyleStr | constants.ToolButtonStyle):
        self.setToolButtonStyle(constants.TOOLBUTTON_STYLE.get_enum_value(style))
        for btn in self._menu_buttons:
            btn.set_style(style)

    def get_style(self) -> constants.ToolButtonStyleStr:
        """Return current style.

        Returns:
            style
        """
        return constants.TOOLBUTTON_STYLE.inverse[self.toolButtonStyle()]

    def add_spacer(self) -> QtGui.QAction:
        spacer = widgets.Widget()
        spacer.set_size_policy("expanding", "expanding")
        return self.addWidget(spacer)

    def set_icon_size(self, size: datatypes.SizeType):
        """Set size of the icons."""
        self.setIconSize(datatypes.to_size(size))

    def get_icon_size(self) -> core.Size:
        return core.Size(self.iconSize())

    def set_font_size(self, size: int):
        with self.edit_font() as font:
            font.set_size(size)

    def is_area_allowed(
        self, area: constants.ToolbarAreaStr | constants.ToolBarArea
    ) -> bool:
        """Check if toolbar is allowed at specified area.

        Args:
            area: area of the toolbar
        """
        return self.isAreaAllowed(constants.TOOLBAR_AREA.get_enum_value(area))

    def set_allowed_areas(self, *areas: constants.ToolbarAreaStr):
        flag = constants.TOOLBAR_AREA.merge_flags(areas)
        self.setAllowedAreas(flag)

    def get_allowed_areas(self) -> list[constants.ToolbarAreaStr]:
        return constants.TOOLBAR_AREA.get_list(self.allowedAreas())

    def get_widgets(self) -> listdelegators.ListDelegator[widgets.QWidget]:
        widgets = [self.widgetForAction(i) for i in self.actions()]
        return listdelegators.ListDelegator(widgets)

    def show_tooltips(
        self,
        value: bool = True,
        duration_ms: int = 2000,
        content: Literal["tool_tip", "shortcut", "status_tip"] = "tool_tip",
    ):
        """Show all the tooltips."""
        if self._tooltip_labels:
            self._tooltip_timer.start(duration_ms)
            return
        for action, i in zip(self.actions(), self.get_widgets()):
            match content:
                case "tool_tip":
                    val = i.toolTip()
                case "shortcut":
                    val = action.shortcut().toString()
                case "status_tip":
                    val = i.statusTip()
                case _:
                    raise ValueError(content)
            if not val:
                continue
            label = widgets.Label(
                val, self, constants.WindowType.ToolTip, alignment="center"
            )
            label.setStyleSheet("border: 1px solid gray;")
            label.hide()
            pos = i.mapToGlobal(core.Point(0, 0))
            label.move(pos)
            label.show()
            self._tooltip_labels.append(label)
        self._tooltip_timer.start(duration_ms)

    def hide_tooltips(self):
        """Hide all the tooltips."""
        for label in self._tooltip_labels:
            label.hide()
            label.deleteLater()
        self._tooltip_labels = []


class ToolBar(ToolBarMixin, widgets.QToolBar):
    """Movable panel that contains a set of controls."""


if __name__ == "__main__":
    app = widgets.app()
    toolbar = ToolBar("test")
    toolbar[0]
    toolbar.add_action(text="test", tool_tip="tesf dsfsdfdsfdsfsdfsdffst")
    toolbar.add_action(text="test2", tool_tip="test2")
    toolbar.add_action(text="test2", tool_tip="test2", shortcut="Ctrl+A")
    radio = widgets.RadioButton("abc")
    action = toolbar.addWidget(radio)
    toolbar.show()
    toolbar.show_tooltips(content="shortcut")
    app.exec()



from __future__ import annotations

from collections import defaultdict
import contextlib

# import inspect
import itertools
import logging
import re
import types
from typing import TYPE_CHECKING, Any, TypeVar, get_args

from prettyqt import constants, core
from prettyqt.utils import classhelpers, datatypes, helpers, listdelegators


if TYPE_CHECKING:
    from collections.abc import Callable, Generator, Sequence

    from prettyqt import eventfilters

T = TypeVar("T", bound=core.QObject)

counter_dict: defaultdict = defaultdict(itertools.count)

logger = logging.getLogger(__name__)

_properties = {}
_signals = {}


def get_properties(klass):
    if klass not in _properties:
        metaobj = core.MetaObject(klass.staticMetaObject)
        _properties[klass] = [i.get_name() for i in metaobj.get_properties()]
    return _properties[klass]


def get_signals(klass):
    if klass not in _signals:
        metaobj = core.MetaObject(klass.staticMetaObject)
        _signals[klass] = [i.get_name() for i in metaobj.get_signals()]
    return _signals[klass]


class ObjectMixin:
    def __init__(self, *args, **kwargs):
        self._eventfilters = set()
        # klass = type(self)
        # if issubclass(klass, core.QObject) and klass not in _properties:
        #     metaobj = core.MetaObject(klass.staticMetaObject)
        #     _properties[klass] = [i.get_name() for i in metaobj.get_properties()]
        #     _signals[klass] = [i.get_name() for i in metaobj.get_signals()]
        new = {}
        if kwargs:
            mapper = self._get_map()
            props = get_properties(type(self))
            signals = get_signals(type(self))
            for k, v in kwargs.items():
                # this allows snake_case naming.
                camel_k = helpers.to_lower_camel(k)
                if camel_k != k and camel_k in kwargs:
                    logger.warning("%s defined twice: %s / %s", k, v, kwargs[camel_k])
                # allow str values instead of enum
                if camel_k in mapper and isinstance(v, str):
                    new[camel_k] = mapper[camel_k][v]
                # allow str values for common icon kwargs
                elif camel_k in {"windowIcon", "icon"} and isinstance(v, str):
                    from prettyqt import iconprovider

                    new[camel_k] = iconprovider.get_icon(v)
                # kwargs which need camel-casing
                elif camel_k in props or camel_k in signals:
                    new[camel_k] = v
                else:
                    new[k] = v
        super().__init__(*args, **new)

    def _get_map(self):
        """Can be implemented by subclasses to support str -> Enum conversion.

        To get data from all subclasses, we always fetch _get_map from super(),
        append our own shit and return it.
        """
        return {}

    def __pretty__(
        self, fmt: Callable[[Any], Any], **kwargs: Any
    ) -> Generator[Any, None, None]:
        yield f"{type(self).__name__}("
        yield 1
        for k, v in self.get_properties(only_writable=True).items():
            yield f"{k}={v!r}"
            yield 0
            for ef in self._eventfilters:
                yield f"Eventfilter={ef.__class__.__name__}"
                yield 0
        yield -1
        yield ")"

    # def __repr__(self):  # we already monkeypatch QObject
    #     return get_repr(self, self.objectName())

    def __setstate__(self, state):
        self.set_properties(state)

    def __getstate__(self):
        return self.get_properties()

    def __reduce__(self):
        return type(self), (), self.__getstate__()

    def __getattr__(self, val):
        cameled = helpers.to_lower_camel(val)
        if cameled in dir(self):
            return getattr(self, cameled)
        raise AttributeError(val)

    def installEventFilter(self, filter_: core.QObject | str, **kwargs):
        """Override to also allow setting eventfilters by name."""
        if filter_ in self._eventfilters:
            logger.warning("Installing same EventFilter multiple times to %s.", self)
            return
        match filter_:
            case core.QObject():
                pass
            case str():
                from prettyqt import eventfilters

                kls = classhelpers.get_class_for_id(eventfilters.BaseEventFilter, filter_)
                filter_ = kls(parent=self, **kwargs)
            case _:
                raise ValueError(filter_)
        self._eventfilters.add(filter_)
        super().installEventFilter(filter_)

    def removeEventFilter(self, eventfilter: core.QObject):
        if eventfilter not in self._eventfilters:
            logger.warning("Trying to remove non-installed EventFilter.")
            return
        self._eventfilters.remove(eventfilter)
        super().removeEventFilter(eventfilter)

    def add_callback_for_event(
        self,
        callback: Callable[[core.QEvent], bool],
        include: core.QEvent.Type | Sequence[core.QEvent.Type] | None = None,
        exclude: core.QEvent.Type | Sequence[core.QEvent.Type] | None = None,
    ) -> eventfilters.EventCatcher:
        """Connect widget events to a callback.

        if include is set, it behaves like a whitelist.
        if exclude is set, it behaves like a blacklist.
        The QEvent is passed to the callback as an argument, and the callback
        needs to return True or False to indicate whether the Event should be filtered.

        Arguments:
            callback: Callback to execute when event is triggered
            include: Events to include
            exclude: Events to exclude
        """
        from prettyqt import eventfilters

        eventfilter = eventfilters.EventCatcher(include, exclude, callback, parent=self)
        self.installEventFilter(eventfilter)
        return eventfilter

    def serialize(self) -> dict[str, Any]:
        return self.get_properties()
        # dct = {}
        # for klass in reversed(inspect.getmro(type(self))):
        #     if "serialize_fields" in klass.__dict__:
        #         data = klass.serialize_fields(self)  # type: ignore
        #         dct |= data
        # return dct

    @contextlib.contextmanager
    def signals_blocked(self):
        """Context manager to temporarily block emitting signals."""
        blocked = self.blockSignals(True)
        yield self
        self.blockSignals(blocked)

    @contextlib.contextmanager
    def signal_blocked(
        self, signal: core.SignalInstance, receiver: Callable | core.SignalInstance
    ):
        """Context manager to temporarily disconnect specific signal."""
        signal.disconnect(receiver)
        yield self
        signal.connect(receiver)

    @contextlib.contextmanager
    def properties_set_to(self, **kwargs):
        """Context manager to temporarily set properties to different values."""
        props = {k: self.property(k) for k in kwargs}
        for k, v in kwargs.items():
            self.setProperty(k, v)
        yield self
        for k, v in props.items():
            self.setProperty(k, v)

    def to_json(self):
        dct = self.__getstate__()
        for k, v in dct.items():
            if isinstance(v, ObjectMixin):
                dct[k] = v.to_json()
        return dct

    def set_unique_id(self):
        """Set unique objectName."""
        class_name = type(self).__name__
        count = next(counter_dict[class_name])
        self.setObjectName(f"{class_name}_{count}")

    @classmethod
    def get_qt_base_class(cls) -> type[core.QObject] | None:
        return classhelpers.get_qt_parent_class(cls)

    @classmethod
    def get_static_metaobject(cls) -> core.MetaObject:
        return core.MetaObject(cls.staticMetaObject)

    def get_metaobject(self) -> core.MetaObject:
        return core.MetaObject(self.metaObject())

    def find_children(
        self,
        typ: type[T] = core.QObject,
        name: str | datatypes.PatternType | None = None,
        recursive: bool = True,
        property_selector: dict[str, datatypes.VariantType | Callable] | None = None,
        only_prettyqt_classes: bool = False,
    ) -> listdelegators.ListDelegator[T]:
        """Find children with given type and name.

        Children can be filtered by passing a property selector dictionary.
        It must contain the property name for keys and either a value which must be set
        or a predicate function which gets the property value as an argument
        and must return True if the child should be included.

        Arguments:
            typ: Subclass of QObject (can also be a UnionType)
            name: ObjectName filter. None includes all.
            recursive: whether to search for children recursively.
            property_selector: dict containing PropertyName -> Value/Predicate pairs.
            only_prettyqt_classes: only include objects with prettyqt superpowers.

        Returns:
            list of QObjects
        """
        if isinstance(name, re.Pattern):
            name = core.RegularExpression(name)
        if recursive:
            flag = constants.FindChildOption.FindChildrenRecursively
        else:
            flag = constants.FindChildOption.FindDirectChildrenOnly
        match typ:
            case types.UnionType():
                objects = [
                    i
                    for t in get_args(typ)
                    for i in self.findChildren(t, name=name, options=flag)
                ]
            case type():
                objects = self.findChildren(typ, name=name, options=flag)
            case _:
                raise TypeError(typ)
        if property_selector:
            objects = [
                o
                for o in objects
                for k, v in property_selector.items()
                if (callable(v) and v(o.property(k)))
                or (not callable(v) and o.property(k) == v)
            ]
        if only_prettyqt_classes:
            objects = [i for i in objects if i.__module__.startswith("prettyqt")]
        return listdelegators.ListDelegator(objects)

    def find_child(
        self,
        typ: type[T] = core.QObject,
        name: str | core.QRegularExpression | None = None,
        recursive: bool = True,
    ) -> T | None:
        """Find a child with given type and name."""
        if recursive:
            flag = constants.FindChildOption.FindChildrenRecursively
        else:
            flag = constants.FindChildOption.FindDirectChildrenOnly
        match typ:
            case types.UnionType():
                return next(
                    (item for item in self.find_children(typ, name, recursive=recursive)),
                    None,
                )

            case _:
                return self.findChild(typ, name, flag)  # type: ignore

    def find_parent(
        self, typ: type[T] = core.QObject, name: str | None = None
    ) -> T | None:
        """Find parent with given type or name."""
        node = self
        while node:
            node = node.parent()
            if isinstance(node, typ) and (name is None or node.objectName() == name):
                return node
        return None

    def start_timer(
        self,
        interval: int | str,
        timer_type: constants.TimerTypeStr = "coarse",
    ) -> int | None:
        """Start a timer and return the timer id, to be used in timerEvent."""
        if isinstance(interval, str):
            interval = helpers.parse_time(interval)
        result = self.startTimer(interval, constants.TIMER_TYPE[timer_type])
        return None if result == 0 else result

    def start_callback_timer(
        self,
        callback: Callable,
        interval: int | str,
        single_shot: bool = False,
        timer_type: constants.TimerTypeStr = "coarse",
    ) -> core.Timer:
        """Start timer and execute callback when timeout reached."""
        interval = helpers.parse_time(interval) if isinstance(interval, str) else interval
        timer = core.Timer(
            self,
            single_shot=single_shot,
            interval=interval,
            timer_type=timer_type,
            timeout=callback,
        )
        timer.start()
        return timer

    def get_properties(
        self,
        include_super: bool = True,
        cast: bool = True,
        only_writable: bool = False,
        only_nonempty: bool = False,
    ) -> dict[str, Any]:
        """Get a dictionary containing all properties and their values."""
        metaobj = self.get_metaobject()
        props = metaobj.get_properties(
            include_super=include_super, only_writable=only_writable
        )
        dct = {}
        for i in props:
            if i.get_name() in ["children", "frameShadow", "state"]:
                continue
            val = i.read(self)
            if only_nonempty:
                match val:
                    case _ if hasattr(val, "isNull"):
                        include = not val.isNull()
                    case _ if hasattr(val, "isEmpty"):
                        include = not val.isEmpty()
                    case _ if hasattr(val, "isValid"):
                        include = val.isValid()
                    case _:
                        include = bool(val)
            else:
                include = True
            if not include:
                continue
            dct[i.name()] = datatypes.make_serializable(val) if cast else val
        return dct

    def set_properties(self, props: dict[str, Any], include_super: bool = True):
        """Set properties from a dictionary."""
        metaobj = self.get_metaobject()
        metaprops = metaobj.get_properties(include_super=include_super)
        for metaprop in metaprops:
            if (name := metaprop.name()) in props:
                value = props[name]
                metaprop.write(self, value)

    def get_dynamic_properties(self) -> dict[str, Any]:
        """Get a dictionary with all dynamic properties."""
        return {
            (k := i.data().decode()): self.property(k)
            for i in self.dynamicPropertyNames()
        }

    def bind_property(self, object_name: str, prop_name: str) -> property:
        def getter(self):
            return self.findChild(self, object_name).property(prop_name)

        def setter(self, value):
            self.findChild(self, object_name).setProperty(prop_name, value)

        return property(getter, setter)


class Object(ObjectMixin, core.QObject):
    """The base class of all Qt objects."""


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    obj = core.Object(object_name="jkjk")
    test = widgets.TableWidget()
    meta = obj.get_metaobject()
    prop = meta.get_property(0)
    with app.debug_mode():
        app.sleep(1)



"""Tests for `prettyqt` package."""

import pathlib

import pytest

from prettyqt import core, gui, quick
from prettyqt.utils import InvalidParamError


def test_quickitem():
    item = quick.QuickItem()
    item.get_children_rect()
    item.get_cursor()
    item.get_flags()
    item.set_transform_origin("bottom")
    with pytest.raises(InvalidParamError):
        item.set_transform_origin("test")
    assert item.get_transform_origin() == "bottom"


def test_quickpainteditem():
    item = quick.QuickPaintedItem()
    item.get_fill_color()
    item.get_texture_size()
    item.set_render_target("framebuffer_object")
    with pytest.raises(InvalidParamError):
        item.set_render_target("test")
    assert item.get_render_target() == "framebuffer_object"


def test_quickview():
    view = quick.QuickView()
    path = pathlib.Path.cwd() / "tests" / "qmltest.qml"
    view.set_source(path)
    assert view.get_source() == path
    assert view.get_status() == "ready"


def test_quickwindow():
    window = quick.QuickWindow()
    window.create_texture_from_image(gui.Image())
    window.grab_window()
    window.get_color()
    window.set_text_render_type("native_text")
    with pytest.raises(InvalidParamError):
        window.set_text_render_type("test")
    assert window.get_text_render_type() == "native_text"
    with window.external_commands():
        pass
    runnable = core.Runnable()
    with pytest.raises(InvalidParamError):
        window.schedule_render_job(runnable, "test")
    window.schedule_render_job(runnable, "before_rendering")



"""A lightweight module handling iconic fonts.

It is designed to provide a simple way for creating gui.QIcons from glyphs.

From a user's viewpoint, the main entry point is the ``IconicFont`` class which
contains methods for loading new iconic fonts with their character map and
methods returning instances of ``QIcon``.

"""

from __future__ import annotations

# Standard library imports
from typing import TYPE_CHECKING, Any

from prettyqt import constants, gui
from prettyqt.iconprovider import chariconengine


if TYPE_CHECKING:
    from collections.abc import Iterable

    from prettyqt.qt import QtCore


_default_options = {
    "color": gui.Color(50, 50, 50),
    "color_disabled": gui.Color(150, 150, 150),
    "opacity": 1.0,
    "scale_factor": 1.0,
}

ICON_KW = [
    "char",
    "on",
    "off",
    "active",
    "selected",
    "disabled",
    "on_active",
    "on_selected",
    "on_disabled",
    "off_active",
    "off_selected",
    "off_disabled",
]

VALID_OPTIONS = [
    "active",
    "selected",
    "disabled",
    "on",
    "off",
    "on_active",
    "on_selected",
    "on_disabled",
    "off_active",
    "off_selected",
    "off_disabled",
    "color",
    "color_on",
    "color_off",
    "color_active",
    "color_selected",
    "color_disabled",
    "color_on_selected",
    "color_on_active",
    "color_on_disabled",
    "color_off_selected",
    "color_off_active",
    "color_off_disabled",
    "animation",
    "offset",
    "scale_factor",
    "rotated",
    "hflip",
    "vflip",
]

COLOR_OPTIONS = {
    gui.QIcon.State.On: {
        gui.QIcon.Mode.Normal: ("color_on", "on"),
        gui.QIcon.Mode.Disabled: ("color_on_disabled", "on_disabled"),
        gui.QIcon.Mode.Active: ("color_on_active", "on_active"),
        gui.QIcon.Mode.Selected: ("color_on_selected", "on_selected"),
    },
    gui.QIcon.State.Off: {
        gui.QIcon.Mode.Normal: ("color_off", "off"),
        gui.QIcon.Mode.Disabled: ("color_off_disabled", "off_disabled"),
        gui.QIcon.Mode.Active: ("color_off_active", "off_active"),
        gui.QIcon.Mode.Selected: ("color_off_selected", "off_selected"),
    },
}


def set_global_defaults(**kwargs: Any):
    """Set global defaults for the options passed to the icon painter."""
    for k, v in kwargs.items():
        if k not in VALID_OPTIONS:
            msg = f"Invalid option {k!r}"
            raise KeyError(msg)
        _default_options[k] = v


class FontError(Exception):
    """Exception for font errors."""


class IconicFont:
    """Main class for managing iconic fonts."""

    def __init__(self, *args):
        super().__init__()
        self.icon_cache = {}
        self.fonts = {font.prefix: font for font in args}

    def paint(
        self,
        painter: gui.Painter,
        rect: QtCore.QRect,
        mode: gui.QIcon.Mode,
        state: gui.QIcon.State,
        options: Iterable[dict[str, Any]],
    ):
        color_str, char = COLOR_OPTIONS[state][mode]
        for opt in options:
            painter.save()
            color = gui.Color(opt[color_str])
            painter.setPen(color)

            # A 16 pixel-high icon yields a font size of 14, which is pixel perfect
            # for font-awesome. 16 * 0.875 = 14
            # The reason why the glyph size is smaller than the icon size is to
            # account for font bearing.

            draw_size = round(0.875 * rect.height() * opt["scale_factor"])
            # Animation setup hook
            if (animation := opt.get("animation")) is not None:
                animation.setup(painter, rect)
            font = self.fonts[opt["prefix"]].get_font(draw_size)
            painter.setFont(font)
            if "offset" in opt:
                rect.translate(
                    round(opt["offset"][0] * rect.width()),
                    round(opt["offset"][1] * rect.height()),
                )
            x_center = rect.width() * 0.5
            y_center = rect.height() * 0.5
            painter.translate(x_center, y_center)
            transform = gui.Transform()
            if opt.get("vflip") is True:
                transform.scale(1, -1)
            if opt.get("hflip") is True:
                transform.scale(-1, 1)
            if "rotated" in opt:
                transform.rotate(opt["rotated"])
            painter.setTransform(transform, True)
            painter.translate(-x_center, -y_center)
            if (opacity := opt.get("opacity")) is not None:
                painter.setOpacity(opacity)

            painter.drawText(rect, int(constants.ALIGN_CENTER), opt[char])  # type: ignore
            painter.restore()

    def has_valid_font_ids(self) -> bool:
        """Validates instance's font ids are loaded to QFontDatabase.

        It is possible that QFontDatabase was reset or QApplication was recreated
        in both cases it is possible that font is not available.
        """
        # Check stored font ids are still available
        return all(font.is_valid() for font in self.fonts.values())

    def icon(self, *names, **kwargs) -> gui.QIcon:
        """Returns a gui.QIcon object corresponding to the provided icon name."""
        cache_key = f"{names}{kwargs}"
        if cache_key in self.icon_cache:
            return self.icon_cache[cache_key]
        opts = kwargs.pop("options", [{}] * len(names))
        if len(opts) != len(names):
            msg = f'"options" must be a list of size {len(names)}'
            raise TypeError(msg)
        parsed_options = [self._parse_options(o, kwargs, n) for o, n in zip(opts, names)]
        engine = chariconengine.CharIconEngine(self, parsed_options)
        icon = gui.QIcon(engine)
        self.icon_cache[cache_key] = icon
        return icon

    def _parse_options(
        self, specific_options: dict, general_options: dict, name: str
    ) -> dict[str, Any]:
        options = dict(_default_options, **general_options) | specific_options
        # Handle icons for modes (Active, Disabled, Selected, Normal)
        # and states (On, Off)
        char = options.get("char", name)
        on = options.get("on", char)
        off = options.get("off", char)
        active = options.get("active", on)
        selected = options.get("selected", active)
        disabled = options.get("disabled", char)
        on_active = options.get("on_active", active)
        on_selected = options.get("on_selected", selected)
        on_disabled = options.get("on_disabled", disabled)
        off_active = options.get("off_active", active)
        off_selected = options.get("off_selected", selected)
        off_disabled = options.get("off_disabled", disabled)

        icon_dict = {
            "char": char,
            "on": on,
            "off": off,
            "active": active,
            "selected": selected,
            "disabled": disabled,
            "on_active": on_active,
            "on_selected": on_selected,
            "on_disabled": on_disabled,
            "off_active": off_active,
            "off_selected": off_selected,
            "off_disabled": off_disabled,
        }
        names = [icon_dict.get(kw, name) for kw in ICON_KW]
        chars = []
        for icon_name in names:
            if "." not in icon_name:
                msg = "Invalid icon name"
                raise Exception(msg)  # noqa: TRY002
            prefix, n = icon_name.split(".")
            if prefix not in self.fonts:
                msg = f"Invalid font prefix {prefix!r}"
                raise Exception(msg)  # noqa: TRY002
            if n not in self.fonts[prefix].charmap:
                msg = f"Invalid icon name {n!r} in font {prefix!r}"
                raise Exception(msg)  # noqa: TRY002
            chars.append(self.fonts[prefix].charmap[n])
        options |= dict(zip(*(ICON_KW, chars)))
        options["prefix"] = prefix

        # Handle colors for modes (Active, Disabled, Selected, Normal)
        # and states (On, Off)
        color = options.get("color")
        options.setdefault("color_on", color)
        options.setdefault("color_active", options["color_on"])
        options.setdefault("color_selected", options["color_active"])
        options.setdefault("color_on_active", options["color_active"])
        options.setdefault("color_on_selected", options["color_selected"])
        options.setdefault("color_on_disabled", options["color_disabled"])
        options.setdefault("color_off", color)
        options.setdefault("color_off_active", options["color_active"])
        options.setdefault("color_off_selected", options["color_selected"])
        options.setdefault("color_off_disabled", options["color_disabled"])
        return options
